"""
File       : MSUnmergedRSE.py
Description: Provides a document Template for the MSUnmerged MicroServices
"""

from pymongo.errors  import NotPrimaryError
# from pymongo.results import results as MongoResults


class MSUnmergedRSE(dict):
    """
    A minimal RSE information representation to serve the needs
    of the MSUnmerged Micro Service.
    """
    def __init__(self, rseName, **kwargs):
        super(MSUnmergedRSE, self).__init__(**kwargs)

        self.rseName = rseName
        self.update(self.defaultDoc())
        self.mongoFilter = {'name': self['name']}

    def defaultDoc(self):
        """
        Returns the data schema for a record in MongoDB.
        :return: A simple dictionary populated with default values
        """
        # NOTE: totalNumFiles reflects the total number of files at the RSE as
        #       fetched from the Rucio Consistency Monitor. Once the relevant
        #       protected paths have been filtered out and the path been cut to the
        #       proper depth (as provided by the WMStats Protected LFNs interface),
        #       then the final number (but on a directory level rather than on
        #       files granularity level) will be put in the counter 'dirsToDelete'

        # NOTE: The type of msUnmergedRSE['files']['toDelete'] is a dictionary of
        #       of generators holding the filters for the files to be deleted e.g.:
        #       msUnmergedRSE['files']['toDelete'] = {
        #          '/store/unmerged/Run2018B/TOTEM42/MINIAOD/22Feb2019-v1': <filter at 0x7f3699d93208>,
        #          '/store/unmerged/Run2018B/TOTEM21/AOD/22Feb2019-v1': <filter at 0x7f3699d93128>,
        #          '/store/unmerged/Run2018D/MuonEG/RAW-RECO/TopMuEG-12Nov2019_UL2018-v1': <filter at 0x7f3699d93668>}
        defaultDoc = {
            "name": self.rseName,
            "pfnPrefix": None,
            "isClean": False,
            "rucioConMonStatus": None,
            "timestamps": {'rseConsStatTime': 0.0,
                           'prevStartTime': 0.0,
                           'startTime': 0.0,
                           'prevEndTime': 0.0,
                           'endTime': 0.0},
            "counters": {"totalNumFiles": 0,
                         "totalNumDirs": 0,
                         "dirsToDelete": 0,
                         "filesToDelete": 0,
                         "filesDeletedSuccess": 0,
                         "filesDeletedFail": 0,
                         "dirsDeletedSuccess": 0,
                         "dirsDeletedFail": 0,
                         "gfalErrors": {}},
            "files": {"allUnmerged": [],
                      "toDelete": {},
                      "protected": {},
                      "deletedSuccess": set(),
                      "deletedFail": set()},
            "dirs": {"allUnmerged": set(),
                     "toDelete": set(),
                     "protected": set(),
                     "deletedSuccess": set(),
                     "deletedFail": set()}
        }
        return defaultDoc

    def buildMongoProjection(self, fullRSEToDB=False):
        """
        Returns the correct mongoProjection based on the `fullRSEToDB` flag passed
        :param fullRSEToDB: Flag to decide whether the whole RSE object is to be
                        included in the projection ergo saved into the database.
        :return:        A dictionary type MongoDB Projection
        """
        mongoProjection = {
            "_id": False,
            "name": True,
            "pfnPrefix": True,
            "rucioConMonStatus": True,
            "isClean" : True,
            "timestamps": True,
            "counters": True,
            "dirs": True}
        if fullRSEToDB:
            mongoProjection.update({"files": True})
        return mongoProjection

    def readRSEFromMongoDB(self, collection, useProjection=False):
        """
        A method to read the RSE object from Database and update it's fields.
        :param collection: The MongoDB collection to read from
        :param useProjection: Use the projection returned by self.buildProjection()
                              while reading from MongoDB
        :return:           True if read and update were both successful, False otherwise.
        """
        if useProjection:
            mongoProjection = self.buildMongoProjection()
            mongoRecord = collection.find_one(self.mongoFilter, projection=mongoProjection)
        else:
            mongoRecord = collection.find_one(self.mongoFilter)

        # update the list fields read from MongoDB back to strictly pythonic `set`s
        if mongoRecord:
            for field in mongoRecord:
                if field == 'dirs' or field == 'files':
                    for fieldKey, fieldList in mongoRecord[field].items():
                        if isinstance(self[field][fieldKey], set):
                            mongoRecord[field][fieldKey] = set(fieldList)
            self.update(mongoRecord)
            return True
        else:
            return False

    def writeRSEToMongoDB(self, collection, fullRSEToDB=False, overwrite=False, retryCount=0):
        """
        A method to write/update the RSE at the Database from the current object.
        :param collection:     The MonogoDB collection to write on
        :param fullRSEToDB:    Bool flag, used to trigger dump of the whole RSE object to
                               the database with the `files` section (excluding the generator objects)
                               NOTE: if fullRSEToDB=False and a previous record for the RSE already exists
                                     the fields missing from the projection won't be updated
                                     during this write operation but will preserver their values.
                                     To completely refresh and RSE record in the database use
                                     self.purgeRSEAtMongoDB first.
        :param overwrite:      A flag to note if the currently existing document into
                               the database is about to be completely replaced or just
                               fields update is to happen.
        :param retryCount:     The number of retries for the write operation if failed due to
                               `NotPrimaryError. Possible values:
                               0   - the write operation will be tried exactly once and no retries will happen
                               > 0 - the write operation will be retried this number of times
                               < 0 - the write operation will never be tried
        :return:               True if update was successful, False otherwise.
        """
        # NOTE: The fields to be manipulated are only those which are compatible
        #       with MongoDB (i.e. here we avoid any field holding a strictly
        #       pythonic objects, like generators etc.)

        updateFields = {}
        mongoProjection = self.buildMongoProjection(fullRSEToDB)
        for field in mongoProjection:
            if mongoProjection[field]:
                # convert back the strictly pythonic `set`s && `generators` to Json compatible lists
                if field == 'dirs':
                    updateFields[field] = {}
                    for dirKey, dirSet in self[field].items():
                        if isinstance(dirSet, set):
                            updateFields[field][dirKey] = list(dirSet)
                elif field == 'files':
                    updateFields[field] = {}
                    for fileKey, fileSet in self[field].items():
                        if isinstance(fileSet, set):
                            updateFields[field][fileKey] = list(fileSet)
                        elif isinstance(fileSet, dict):
                            # Iterating through the filterNames here, and recording only empty lists for filter values
                            # NOTE: We can either execute the filter and write every single file in the database
                            #       or if we need it we may simply use the filterName to recreate it.
                            updateFields[field][fileKey] = dict([(filterName, []) for filterName in list(self[field][fileKey])])
                        else:
                            updateFields[field][fileKey] = self[field][fileKey]
                else:
                    updateFields[field] = self[field]

        updateOps = {'$set': updateFields}
        # NOTE: NotPrimaryError is a recoverable error, caused by a session to a
        #       non primary backend part of a replicaset.
        while retryCount >= 0:
            try:
                if overwrite:
                    result = collection.replace_one(self.mongoFilter, updateFields, upsert=True)
                else:
                    result = collection.update_one(self.mongoFilter, updateOps, upsert=True)
                break
            except NotPrimaryError:
                if retryCount:
                    # msg = "Failed write operation to MongoDB. %s retries left."
                    # self.logger.warning(msg, retryCount)
                    retryCount -= 1
                else:
                    # msg = "Failed write operation to MongoDB. All retries exhausted."
                    # self.logger.warning(msg, retryCount)
                    raise

        # NOTE: If and `upsert` took place the modified_count for both operations is 0
        #       because no modification took place  but rather an insertion.
        if result.modified_count or result.upserted_id:
            return True
        else:
            return False

    def resetRSE(self, collection, keepTimestamps=False, keepCounters=False, retryCount=0):
        """
        Resets all records of the RSE object to default values  and write the
        document to MongoDB
        :param keepTimestamps: Bool flag to keep the timestamps
        :param keepCounters:   Bool flag to keep the counters
        :param retryCount:     The number of retries for the write operation if failed due to `NotPrimaryError.
        :return:               True if operation succeeded.
        """
        resetDoc = self.defaultDoc()

        if keepTimestamps:
            resetDoc['timestamps'] = self['timestamps']
        if keepCounters:
            resetDoc['counters'] = self['counters']

        self.update(resetDoc)
        writeResult = self.writeRSEToMongoDB(collection, fullRSEToDB=True, overwrite=True, retryCount=retryCount)
        return writeResult



#!/bin/env python

"""
WorkerNode unittest for WMRuntime/WMSpec

"""
from __future__ import print_function

import os
import os.path
import random
import shutil
# Basic libraries
import unittest

import sys
# from WMCore.WMSpec.StdSpecs.ReReco  import rerecoWorkload, getTestArguments
from WMCore_t.WMSpec_t.TestSpec import createTestWorkload

import WMCore.WMRuntime.Bootstrap as Bootstrap
# DataStructs
from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.JobPackage import JobPackage
from WMCore.DataStructs.Subscription import Subscription
from WMCore.FwkJobReport.Report import Report
# Factories
from WMCore.JobSplitting.Generators.GeneratorFactory import GeneratorFactory
from WMCore.JobSplitting.SplitterFactory import SplitterFactory
# Misc WMCore
from WMCore.Services.UUIDLib import makeUUID
# WMRuntime
from WMCore.WMRuntime.Unpacker import runUnpacker as RunUnpacker
# Builders
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
# Init junk
from WMQuality.TestInit import TestInit


def getListOfTasks(workload):
    """
    _getListOfTasks_

    Get a complete list of tasks in a workload
    Returns a list of task helpers
    """
    listOfTasks = []

    for primeTask in workload.taskIterator():
        for task in primeTask.taskIterator():
            listOfTasks.append(task)

    return listOfTasks


def miniStartup(thisDir=os.getcwd()):
    """
    This is an imitation of the startup script
    I don't want to try sourcing the main() of the startup
    Or run this in subprocess

    """
    Bootstrap.setupLogging(thisDir)
    job = Bootstrap.loadJobDefinition()
    task = Bootstrap.loadTask(job)
    Bootstrap.createInitialReport(job=job,
                                  reportName="Report.0.pkl")
    monitor = Bootstrap.setupMonitoring(logName="Report.0.pkl")

    task.build(thisDir)
    task.execute(job)

    task.completeTask(jobLocation=os.path.join(thisDir, 'WMTaskSpace'),
                      reportName="Report.0.pkl")

    if monitor.is_alive():
        monitor.shutdown()

    return


class RuntimeTest(unittest.TestCase):
    """
    _RuntimeTest_

    A unittest to test the WMRuntime/WMSpec/Storage/etc tree
    """

    def setUp(self):
        """
        Basic setUp

        """

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()

        self.testDir = self.testInit.generateWorkDir()

        # Random variables
        self.workloadDir = None
        self.unpackDir = None
        self.initialDir = os.getcwd()
        self.origPath = sys.path
        self.thisDirPath = os.path.dirname(__file__)

        # Create some dirs
        os.makedirs(os.path.join(self.testDir, 'packages'))

        return

    def tearDown(self):
        """
        _tearDown_

        Remove any references you put directly into the modules
        """

        self.testInit.delWorkDir()

        # Clean up imports
        if 'WMSandbox' in sys.modules.keys():
            del sys.modules['WMSandbox']
        if 'WMSandbox.JobIndex' in sys.modules.keys():
            del sys.modules['WMSandbox.JobIndex']

        return

    def setupTestWorkload(self, workloadName='Test', emulator=True):
        """
        Creates a test workload for us to run on, hold the basic necessities.
        """

        workloadDir = os.path.join(self.testDir, workloadName)

        # arguments = getTestArguments()

        # workload = rerecoWorkload("Tier1ReReco", arguments)
        # rereco = workload.getTask("ReReco")

        workload = createTestWorkload(emulation=emulator)
        rereco = workload.getTask("ReReco")

        # Set environment and site-local-config
        siteConfigPath = os.path.join(workloadDir, 'SITECONF/local/JobConfig/')
        if not os.path.exists(siteConfigPath):
            os.makedirs(siteConfigPath)
        shutil.copy(os.path.join(self.thisDirPath, 'site-local-config.xml'), siteConfigPath)
        shutil.copy(os.path.join(self.thisDirPath, 'storage.json'), os.path.join(siteConfigPath, '..'))
        environment = rereco.data.section_('environment')
        environment.CMS_PATH = workloadDir
        environment.SITECONFIG_PATH = os.path.join(workloadDir, 'SITECONF/local')

        taskMaker = TaskMaker(workload, workloadDir)
        taskMaker.skipSubscription = True
        taskMaker.processWorkload()

        workload.save(workloadName)

        return workload

    def unpackComponents(self, workload):
        """
        Run the unpacker to build the directories
        IMPORTANT NOTE:
          This is not how we do things on the worker node
          On the worker node we do not run multiple tasks
          So here we create multiple tasks in different directories
          To mimic running on multiple systems

        """

        listOfTasks = getListOfTasks(workload=workload)

        self.unpackDir = os.path.join(self.testDir, 'unpack')

        if not os.path.exists(self.unpackDir):
            os.makedirs(self.unpackDir)

        os.chdir(self.unpackDir)

        sandbox = workload.data.sandbox

        for task in listOfTasks:
            # We have to create a directory, unpack in it, and then get out
            taskName = task.name()
            taskDir = os.path.join(self.unpackDir, taskName)
            if not os.path.exists(taskDir):
                # Well then we have to make it
                os.makedirs(taskDir)
            os.chdir(taskDir)
            # Now that we're here, run the unpacker

            package = os.path.join(self.testDir, 'packages', '%sJobPackage.pkl' % (taskName))
            jobIndex = 1

            RunUnpacker(sandbox=sandbox, package=package,
                        jobIndex=jobIndex, jobname=taskName)

            # And go back to where we started
            os.chdir(self.unpackDir)

        os.chdir(self.initialDir)

        return

    def createWMBSComponents(self, workload):
        """
        Create the WMBS Components for this job

        """

        listOfTasks = []

        for primeTask in workload.taskIterator():
            for task in primeTask.taskIterator():
                listOfTasks.append(task)

        for task in listOfTasks:
            fileset = self.getFileset()
            sub = self.createSubscriptions(task=task,
                                           fileset=fileset)

        return

    def createSubscriptions(self, task, fileset):
        """
        Create a subscription based on a task


        """
        taskType = task.taskType()
        work = task.makeWorkflow()

        sub = Subscription(fileset=fileset,
                           workflow=work,
                           split_algo="FileBased",
                           type=taskType)

        package = self.createWMBSJobs(subscription=sub,
                                      task=task)

        packName = os.path.join(self.testDir, 'packages',
                                '%sJobPackage.pkl' % (task.name()))
        package.save(packName)

        return sub

    def createWMBSJobs(self, subscription, task):
        """
        Create the jobs for WMBS Components
        Send a subscription/task, get back a package.

        """

        splitter = SplitterFactory()
        geneFac = GeneratorFactory()
        jobfactory = splitter(subscription=subscription,
                              package="WMCore.DataStructs",
                              generators=geneFac.makeGenerators(task))
        params = task.jobSplittingParameters()
        jobGroups = jobfactory(**params)

        jobID = 1
        package = JobPackage()
        for group in jobGroups:
            for job in group.jobs:
                job['id'] = jobID
                jobID += 1
                package[job['id']] = job

        return package

    def getFileset(self):
        """
        Get a fileset based on the task

        """

        fileset = Fileset(name='Merge%s' % (type))

        for i in range(0, random.randint(15, 25)):
            # Use the testDir to generate a random lfn
            inpFile = File(lfn="%s/%s.root" % (self.testDir, makeUUID()),
                           size=random.randint(200000, 1000000),
                           events=random.randint(1000, 2000))
            inpFile.setLocation('Megiddo')
            fileset.addFile(inpFile)

        return fileset

    def runJobs(self, workload):
        """
        This might actually run the job.  Who knows?


        """
        listOfTasks = []

        for primeTask in workload.taskIterator():
            listOfTasks.append(primeTask)
            # Only run primeTasks for now

        for task in listOfTasks:
            jobName = task.name()
            taskDir = os.path.join(self.unpackDir, jobName, 'job')
            os.chdir(taskDir)
            sys.path.append(taskDir)

            # Scream, run around in panic, blow up machine
            print("About to run jobs")
            print(taskDir)
            # SITECONFIG_PATH is not available here so set it up so that site config can be loaded in Bootstrap.createInitialReport inside miniStartup
            os.environ['SITECONFIG_PATH'] = os.path.realpath(
                os.path.join(taskDir, '../../../basicWorkload/SITECONF/local'))
            miniStartup(thisDir=taskDir)
            # When exiting, go back to where you started
            os.chdir(self.initialDir)
            sys.path.remove(taskDir)

        return

    def testA_CreateWorkload(self):
        """
        _CreateWorkload_

        Create a workload
        Unpack the workload
        Check for consistency
        """

        workloadName = 'basicWorkload'
        workload = self.setupTestWorkload(workloadName=workloadName)

        self.createWMBSComponents(workload=workload)

        taskNames = []
        for task in getListOfTasks(workload=workload):
            taskNames.append(task.name())

        workloadPath = os.path.join(self.testDir, workloadName, "TestWorkload")
        siteConfigDir = os.path.join(self.testDir, workloadName, 'SITECONF/local/JobConfig/')

        # Pre-run checks

        # Does it have the right directories?
        dirList = os.listdir(workloadPath)
        self.assertCountEqual(dirList, ['WMSandbox', 'TestWorkload-Sandbox.tar.bz2'])
        dirList = os.listdir(os.path.join(workloadPath, 'WMSandbox'))
        for taskName in taskNames:
            self.assertTrue(taskName in dirList)

        # Do we have job packages
        for task in taskNames:
            self.assertTrue('%sJobPackage.pkl' % (task) in os.listdir(os.path.join(self.testDir, 'packages')))

        # Does it have the SITECONF?
        self.assertTrue('site-local-config.xml' in os.listdir(siteConfigDir))

        # Now actually see if you can unpack it.
        self.unpackComponents(workload=workload)

        # Check for proper unpacking
        # Check the the task has the right directories,
        # and that the PSetTweaks and WMSandbox directories
        # have the right contents
        taskContents = ['WMSandbox', 'WMCore', 'PSetTweaks']
        PSetContents = ['PSetTweak.pyc', 'CVS', 'PSetTweak.py',
                        '__init__.pyc', 'WMTweak.py', '__init__.py']
        taskSandbox = ['JobPackage.pcl', 'JobIndex.py', '__init__.py', 'WMWorkload.pkl']
        taskSandbox.extend(taskNames)  # Should have a directory for each task

        for task in taskNames:
            self.assertTrue(task in os.listdir(os.path.join(self.testDir, 'unpack')))
            taskDir = os.path.join(self.testDir, 'unpack', task, 'job')
            self.assertTrue(os.path.isdir(taskDir))
            self.assertEqual(os.listdir(taskDir).sort(), taskContents.sort())
            self.assertEqual(os.listdir(os.path.join(taskDir, 'WMSandbox')).sort(),
                             taskSandbox.sort())
            self.assertEqual(os.listdir(os.path.join(taskDir, 'PSetTweaks')).sort(),
                             PSetContents.sort())

        # And we're done.
        # Assume if we got this far everything is good

        # At the end, copy the directory
        # if os.path.exists('tmpDir'):
        #    shutil.rmtree('tmpDir')
        # shutil.copytree(self.testDir, 'tmpDir')

        return

    def testB_EmulatorTest(self):
        """
        _EmulatorTest_

        This is where things get scary.  We need to not only unpack the job,
        but also ascertain whether it can run locally in emulator mode.

        This requires...uh...emulator emulation.
        """

        # Assume all this works, because we tested it in testA
        workloadName = 'basicWorkload'
        workload = self.setupTestWorkload(workloadName=workloadName)

        self.createWMBSComponents(workload=workload)

        self.unpackComponents(workload=workload)

        self.runJobs(workload=workload)

        # Check the report
        taskDir = os.path.join(self.testDir, 'unpack/ReReco/job/WMTaskSpace')
        report = Report()
        report.load(os.path.join(taskDir, 'Report.0.pkl'))
        cmsReport = report.data.cmsRun1

        # Now validate the report
        self.assertEqual(report.getSiteName(), 'T1_US_FNAL')
        # self.assertEqual(report.data.hostName, socket.gethostname())
        self.assertTrue(report.data.completed)

        # Should have status 0 (emulator job)
        self.assertEqual(cmsReport.status, 0)

        # Should have one output module
        self.assertEqual(cmsReport.outputModules, ['TestOutputModule'])

        # It should have one file for input and output
        self.assertEqual(cmsReport.input.PoolSource.files.fileCount, 1)
        self.assertEqual(cmsReport.output.TestOutputModule.files.fileCount, 1)

        # So, um, I guess we're done

        # At the end, copy the directory
        # if os.path.exists('tmpDir'):
        #    shutil.rmtree('tmpDir')
        # shutil.copytree(self.testDir, 'tmpDir')

        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_List_

MySQL implementation of ListFileset

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class List(DBFormatter):
    sql = "select id, name, open, last_update from wmbs_fileset order by last_update, name"

    def format(self, result):
        """
        Some standardised formatting
        """
        out = []
        for r in result:
            for i in r.fetchall():
                res = i
                j = int(i[0]), i[1], self.truefalse(i[2]), i[3]
                out.append(j)
        return out



"""
WorkQueueElementResult

A dictionary based object meant to represent a WorkQueue block
"""
from __future__ import division

#Can we re-use WorkQueueElement for this?




class WorkQueueElementResult(dict):
    """Class to hold the status of a related group of WorkQueueElements"""
    def __init__(self, **kwargs):
        dict.__init__(self)
        self.update(kwargs)

        self.setdefault('WMSpec', None)
        self.setdefault('Elements', [])
        self.setdefault('Status', self.status())
        if self['Elements']:
            self.setdefault('EventsWritten',
                            sum([x['EventsWritten'] for x in self['Elements']]))
            self.setdefault('FilesProcessed',
                            sum([x['FilesProcessed'] for x in self['Elements']]))
            self.setdefault('Jobs', sum([x['Jobs'] for x in self['Elements']]))
            self.setdefault('PercentComplete',
                            int(sum([x['PercentComplete'] for x in self['Elements']],
                                0.0) / len(self['Elements'])))
            self.setdefault('PercentSuccess',
                            int(sum([x['PercentSuccess'] for x in self['Elements']],
                                0.0) / len(self['Elements'])))
            self.setdefault('RequestName', self['Elements'][0]['RequestName'])
            self.setdefault('TeamName', self['Elements'][0]['TeamName'])
            self.setdefault('Priority', self['Elements'][0]['Priority'])
            self.setdefault('ParentQueueId', self['Elements'][0]['ParentQueueId'])
        elif self.get('ParentQueueElement'):
            self.setdefault('EventsWritten', self['ParentQueueElement']['EventsWritten'])
            self.setdefault('FilesProcessed', self['ParentQueueElement']['FilesProcessed'])
            self.setdefault('Jobs', self['ParentQueueElement']['Jobs'])
            self.setdefault('PercentComplete', self['ParentQueueElement']['PercentComplete'])
            self.setdefault('PercentSuccess', self['ParentQueueElement']['PercentSuccess'])
            self.setdefault('RequestName', self['ParentQueueElement']['RequestName'])
            self.setdefault('TeamName', self['ParentQueueElement']['TeamName'])
            self.setdefault('Priority', self['ParentQueueElement']['Priority'])
            self.setdefault('ParentQueueId', self['ParentQueueElement'].id)
        else:
            raise RuntimeError("Can create WQEResult: No elements or parent provided")

        # some cross checks
        for i in self['Elements']:
            assert(i['RequestName'] == self['RequestName'])
            assert(i['TeamName'] == self['TeamName'])

    def fractionComplete(self):
        """Return fraction successful"""
        return len(self.completeItems()) / float(len(self['Elements']))

    def completeItems(self):
        """Return complete items"""
        return [x for x in self['Elements'] if x.isComplete()]

    def failedItems(self):
        """Return failed items"""
        return [x for x in self['Elements'] if x.isFailed()]

    def runningItems(self):
        """Return acquired items"""
        return [x for x in self['Elements'] if x.isRunning()]

    def acquiredItems(self):
        """Return available items"""
        return [x for x in self['Elements'] if x.isAcquired()]

    def availableItems(self):
        """Return available items"""
        return [x for x in self['Elements'] if x.isAvailable()]

    def canceledItems(self):
        """Return canceled items"""
        return [x for x in self['Elements'] if x.isCanceled()]

    def cancelRequestedItems(self):
        """Return CancelRequested Items"""
        return [x for x in self['Elements'] if x.isCancelRequested()]

    def status(self):
        """Compute status of elements"""
        if not self['Elements'] and self.get('ParentQueueElement'):
            return self['ParentQueueElement']['Status']
        elif not self['Elements']:
            return None

        if not self.inEndState():
            if self.cancelRequestedItems():
                return 'CancelRequested'
            elif self.runningItems():
                return 'Running'
            elif self.acquiredItems() or self.availableItems():
                # available in local queue is acquired to parent
                return 'Acquired'
        else:
            if self.canceledItems():
                return "Canceled"
            elif self.failedItems():
                return "Failed"

        # if all elements have same status take that
        status = self['Elements'][0]['Status']
        for element in self['Elements']:
            if element['Status'] != status:
                msg = "Unable to compute overall status of elements: %s"
                raise RuntimeError(msg % str(self['Elements']))
        return status

    def inEndState(self):
        """Are all requests complete (either success or failure)"""
        if 'Status' in self:
            return self['Status'] in ('Done', 'Failed', 'Canceled')
        for element in self['Elements']:
            if not element.inEndState():
                return False
        return True

    def statusMetrics(self):
        """Returns the status & performance metrics"""
        keys = ['Status', 'PercentComplete', 'PercentSuccess']
        return self.fromkeys(keys)

    def formatForWire(self):
        """Format used to send data over network
        """
        to_remove = ['Elements', 'WMSpec']
        result = dict(self)
        for item in to_remove:
            result.pop(item)
        return result

    def getMaxJobElement(self):
        maxJobElement = self['Elements'][0]
        for x in self['Elements']:
            if x['Jobs'] > maxJobElement['Jobs']:
                maxJobElement = x
        return maxJobElement

    def getMinJobElement(self):
        minJobElement = self['Elements'][0]
        for x in self['Elements']:
            if x['Jobs'] < minJobElement['Jobs']:
                minJobElement = x
        return minJobElement



#!/usr/bin/env python
"""
_MultiMergeProductionWorkload_

Sample/Tester for a production workflow that writes several datasets
and associated merges for each dataset.

Production task produces three output datasets,
Merge tasks are used to merge those datasets

"""
from WMCore.WMSpec.WMWorkload import newWorkload
from WMCore.WMSpec.WMStep import makeWMStep
from WMCore.WMSpec.Steps.StepFactory import getStepTypeHelper

#  //
# // Set up the basic workload task and step structure
#//
workload = newWorkload("MultiOutputs")
workload.setOwner("WMTest")
workload.setStartPolicy('MonteCarlo')
workload.setEndPolicy('SingleShot')
workload.setSiteWhitelist(["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"])

#  //
# // set up the production task
#//
production = workload.newTask("Production")
#WARNING: this is arbitrary task type (wmbs schema only supprot "Processing", "Merge", "Harvest") - maybe add "MCProduction"
production.setTaskType("Merge")
production.addProduction(totalEvents = 1000)
prodCmssw = production.makeStep("cmsRun1")
prodCmssw.setStepType("CMSSW")
prodStageOut = prodCmssw.addStep("stageOut1")
prodStageOut.setStepType("StageOut")
production.applyTemplates()
production.setSplittingAlgorithm("EventBased", events_per_job = 100)


#  //
# // populate the details of the production tasks
#//
#  //
# // production cmssw step
#//
#
# TODO: Anywhere helper.data is accessed means we need a method added to the
# type based helper class to provide a clear API.
prodCmsswHelper = prodCmssw.getTypeHelper()

prodCmsswHelper.data.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
prodCmsswHelper.data.application.setup.softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"

prodCmsswHelper.data.application.configuration.configCacheUrl = "http://whatever"

prodCmsswHelper.addOutputModule("writeOutput1", primaryDataset = "Primary",
                                processedDataset = "Processed1-unmerged",
                                dataTier = "TIERONE")
prodCmsswHelper.addOutputModule("writeOutput2", primaryDataset = "Primary",
                                processedDataset = "Processed2-unmerged",
                                dataTier = "TIERTWO")
prodCmsswHelper.addOutputModule("writeOutput3", primaryDataset = "Primary",
                                processedDataset = "Processed3-unmerged",
                                dataTier = "TIERTHREE")
#print prodCmsswHelper.data



#  //
# // set up the merge task for output dataset 1
#//
merge1 = production.addTask("MergeOutput1")
merge1.setTaskType("Merge")
merge1Cmssw = merge1.makeStep("cmsRun1")
merge1Cmssw.setStepType("CMSSW")
merge1StageOut = merge1Cmssw.addStep("stageOut1")
merge1StageOut.setStepType("StageOut")
merge1.applyTemplates()
merge1.setSplittingAlgorithm("MergeBySize", merge_size = 20000000)

#  //
# // set up the merge task for output dataset 1
#//
merge2 = production.addTask("MergeOutput2")
merge2.setTaskType("Merge")
merge2Cmssw = merge2.makeStep("cmsRun1")
merge2Cmssw.setStepType("CMSSW")
merge2StageOut = merge2Cmssw.addStep("stageOut1")
merge2StageOut.setStepType("StageOut")
merge2.applyTemplates()
merge2.setSplittingAlgorithm("MergeBySize", merge_size = 20000000)

#  //
# // set up the merge task for output dataset 3
#//
merge3 = production.addTask("MergeOutput3")
merge3.setTaskType("Merge")
merge3Cmssw = merge3.makeStep("cmsRun1")
merge3Cmssw.setStepType("CMSSW")
merge3StageOut = merge3Cmssw.addStep("stageOut1")
merge3StageOut.setStepType("StageOut")
merge3.applyTemplates()
merge3.setSplittingAlgorithm("MergeBySize", merge_size = 20000000)



#  //
# // production stage out step
#//
prodStageOutHelper = prodStageOut.getTypeHelper()


#  //
# // merge cmssw step
#//
# point it at the input step from the previous task
merge1.setInputReference(prodCmssw, outputModule = "writeOutput1")
merge2.setInputReference(prodCmssw, outputModule = "writeOutput2")
merge3.setInputReference(prodCmssw, outputModule = "writeOutput3")


#  //
# // populate the details of the merge tasks
#//
merge1CmsswHelper = merge1Cmssw.getTypeHelper()
merge1CmsswHelper.data.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
merge1CmsswHelper.data.application.setup.softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"
merge1CmsswHelper.data.application.configuration.configCacheUrl = "http://whatever"
merge1CmsswHelper.addOutputModule("mergeData1", primaryDataset = "Primary",
                                 processedDataset = "Processed1",
                                 dataTier = "TIER1")

merge2CmsswHelper = merge2Cmssw.getTypeHelper()
merge2CmsswHelper.data.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
merge2CmsswHelper.data.application.setup.softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"
merge2CmsswHelper.data.application.configuration.configCacheUrl = "http://whatever"
merge2CmsswHelper.addOutputModule("mergeData2", primaryDataset = "Primary",
                                 processedDataset = "Processed2",
                                 dataTier = "TIER2")


merge3CmsswHelper = merge3Cmssw.getTypeHelper()
merge3CmsswHelper.data.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
merge3CmsswHelper.data.application.setup.softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"
merge3CmsswHelper.data.application.configuration.configCacheUrl = "http://whatever"
merge3CmsswHelper.addOutputModule("mergeData3", primaryDataset = "Primary",
                                 processedDataset = "Processed3",
                                 dataTier = "TIER3")




#print workload.data



#!/usr/bin/env python
from __future__ import (print_function, division)
import unittest
from pprint import pprint
from WMCore.Services.WMStats.WMStatsReader import WMStatsReader
from WMCore.Services.RequestDB.RequestDBWriter import RequestDBWriter
from WMQuality.TestInitCouchApp import TestInitCouchApp
from WMCore.WMSpec.WMWorkload import newWorkload
from WMCore.Services.WMStats.DataStruct.RequestInfoCollection \
    import RequestInfoCollection, RequestInfo

from WMCore_t.Services_t.WMStats_t.WMStatsDocGenerator \
   import generate_reqmgr_schema, sample_request_info, sample_complete

class WMStatsTest(unittest.TestCase):
    """
    """
    def setUp(self):
        """
        _setUp_
        """
        self.schema = []
        self.couchApps = ["WMStats"]
        self.testInit = TestInitCouchApp('WorkQueueServiceTest')
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = self.schema,
                                useDefault = False)
        dbName = 'wmstats_t'
        self.testInit.setupCouch(dbName, "WMStats", "WMStatsErl")
        reqDBName = "reqmgrdb_t"
        self.testInit.setupCouch(reqDBName, "ReqMgr")
        wmstatsURL = "%s/%s" % (self.testInit.couchUrl, dbName)
        reqDBURL = "%s/%s" % (self.testInit.couchUrl, reqDBName)
        self.reqDBWriter = RequestDBWriter(reqDBURL)
        self.wmstatsReader = WMStatsReader(wmstatsURL,  appName="WMStatsErl", reqdbURL=reqDBURL)
        self.wmstatsReader.defaultStale = {}
        self.wmstatsReader.reqDB.defaultStale = {}
        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """
        self.testInit.tearDownCouch()

    def testWMStatsWriter(self):
        # test getWork
        schema = generate_reqmgr_schema()

        result = self.reqDBWriter.insertGenericRequest(schema[0])
        self.assertEqual(result[0]['ok'], True, 'insert fail')

        result = self.reqDBWriter.updateRequestStatus(schema[0]['RequestName'], "failed")
        self.assertEqual(result, 'OK', 'update fail')

        result = self.reqDBWriter.updateRequestStatus("not_exist_schema", "assigned")
        self.assertEqual(result,'Error: document not found')

        result = self.reqDBWriter.updateRequestProperty(schema[0]['RequestName'], {"Team": 'teamA'})
        self.assertEqual(result, 'OK', 'update fail')

        result = self.reqDBWriter.updateRequestProperty("not_exist_schema", {"Team": 'teamA'})
        self.assertEqual(result, 'Error: document not found')

        totalStats = {'TotalEstimatedJobs': 100, 'TotalInputEvents': 1000, 'TotalInputLumis': 1234, 'TotalInputFiles': 5}
        result = self.reqDBWriter.updateRequestProperty(schema[0]['RequestName'], totalStats)
        self.assertEqual(result, 'OK', 'update fail')

        result = self.reqDBWriter.updateRequestProperty(schema[0]['RequestName'], totalStats)
        self.assertEqual(result, 'OK', 'update fail')

        result = self.reqDBWriter.updateRequestProperty("not_exist_schema", totalStats)
        self.assertEqual(result, 'Error: document not found')

        spec1 = newWorkload(schema[0]['RequestName'])
        production = spec1.newTask("Production")
        production.setTaskType("Merge")
        production.setSiteWhitelist(['TEST_SITE'])
        properties = {"RequestPriority": spec1.priority(),
                      'SiteWhitelist': spec1.getTopLevelTask()[0].siteWhitelist(),
                      'OutputDatasets': spec1.listOutputDatasets()}
        result = self.reqDBWriter.updateRequestProperty(spec1.name(), properties)
        self.assertEqual(result, 'OK', 'update fail')

        spec2 = newWorkload("not_exist_schema")
        production = spec2.newTask("Production")
        production.setTaskType("Merge")
        properties = {"RequestPriority": spec2.priority(),
                      'SiteWhitelist': spec2.getTopLevelTask()[0].siteWhitelist(),
                      'OutputDatasets': spec2.listOutputDatasets()}
        result = self.reqDBWriter.updateRequestProperty(spec2.name(), properties)
        self.assertEqual(result, 'Error: document not found')

        requests = self.wmstatsReader.getRequestByStatus(["failed"], jobInfoFlag = False, legacyFormat = True)
        self.assertEqual(list(requests), [schema[0]['RequestName']])

        requestCollection = RequestInfoCollection(requests)
        result = requestCollection.getJSONData()
        self.assertEqual(list(result), [schema[0]['RequestName']])

        requests = self.wmstatsReader.getActiveData()
        self.assertEqual(list(requests), [schema[0]['RequestName']])
        requests = self.wmstatsReader.getRequestByStatus(["failed"])
        self.assertEqual(list(requests), [schema[0]['RequestName']])

        requests = self.wmstatsReader.getRequestSummaryWithJobInfo(schema[0]['RequestName'])
        self.assertEqual(list(requests), [schema[0]['RequestName']])

    def testCompletedCheck(self):
        self.assertEqual(RequestInfo(sample_request_info).isWorkflowFinished(), False)
        self.assertEqual(RequestInfo(sample_complete).isWorkflowFinished(), True)

if __name__ == '__main__':

    unittest.main()



#!/usr/bin/env python
"""
_TestImpl_

Couple of test implementations for unittests

"""

import os
import os.path
import shutil


from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.Storage.StageOutError import StageOutFailure

class TestWinImpl(StageOutImplV2):
    """
    _WinImpl_

    Test plugin that always returns success

    """
    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum  ):
        return "WIN!!!"

    def doDelete(self, lfn, pfn, pnn, command, options, protocol  ):
        pass



#!/usr/bin/env python
"""
Unittests for IteratorTools functions
"""
import unittest

from WMCore.ReqMgr.Utils.AuxValidation import validateUnifiedConfig
from WMCore.REST.Error import InvalidUnifiedSchema


class AuxValidationTests(unittest.TestCase):
    """
    Unittests for ReqMgr Utils Validation functions
    """

    def test1ValidateUnifiedConfig(self):
        """
        Test the `validateUnifiedConfig` function with wrong input data
        """
        # bad data type
        for wrongData in ["", [], 123, {}]:
            with self.assertRaises(RuntimeError):
                validateUnifiedConfig(wrongData)

        # missing top level parameters
        userD = {"tiers_to_DDM": {"value": ["LHE"], "description": "tiers to DDM"}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)
        userD = {"tiers_to_DDM": {"value": ["LHE"], "description": "tiers to DDM"},
                 "tiers_no_DDM": {"value": ["GEN-SIM"], "description": "tiers no DDM"}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)
        userD = {"tiers_to_DDM": {"value": ["LHE"], "description": "tiers to DDM"},
                 "tiers_with_no_custodial": {"value": ["DQM"], "description": "tiers no custodial"}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)

        # and now with 1 unsupported parameter
        userD = {"tiers_to_DDM": {"value": ["LHE"], "description": "tiers to DDM"},
                 "tiers_no_DDM": {"value": ["GEN-SIM"], "description": "tiers no DDM"},
                 "tiers_with_no_custodial": {"value": ["DQM"], "description": "tiers no custodial"},
                 "BAD_KEY": {"value": ["LHE"], "description": "unsupported parameter"}
                 }
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)

    def test2ValidateUnifiedConfig(self):
        """
        Test the `validateUnifiedConfig` function with wrong
        nested schema, but also with the correct and expected schema.
        """
        # missing nested keys
        userD = {"tiers_to_DDM": {},
                 "tiers_no_DDM": {},
                 "tiers_with_no_custodial": {}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)
        userD = {"tiers_to_DDM": {"value": ["LHE"]},
                 "tiers_no_DDM": {"value": ["GEN-SIM"]},
                 "tiers_with_no_custodial": {"value": ["DQM"]}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)
        userD = {"tiers_to_DDM": {"description": "tiers to DDM"},
                 "tiers_no_DDM": {"description": "tiers no DDM"},
                 "tiers_with_no_custodial": {"description": "tiers no custodial"}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)

        # now with too many nested keys
        userD = {"tiers_to_DDM": {"value": ["LHE"], "description": "tiers to DDM", "BAD_KEY": "blah"},
                 "tiers_no_DDM": {"value": ["GEN-SIM"], "description": "tiers no DDM"},
                 "tiers_with_no_custodial": {"value": ["DQM"], "description": "tiers no custodial"}}
        with self.assertRaises(RuntimeError):
            validateUnifiedConfig(userD)

        # last but not least, a good looking and supported schema
        userD = {"tiers_to_DDM": {"value": ["LHE"], "description": "tiers to DDM"},
                 "tiers_no_DDM": {"value": ["GEN-SIM"], "description": "tiers no DDM"},
                 "tiers_with_no_custodial": {"value": ["DQM"], "description": "tiers no custodial"}}
        validateUnifiedConfig(userD)

    def testInvalidUnifiedSchemaException(self):
        """
        Test the `InvalidUnifiedSchema` exception
        """
        errorMsg = ""
        try:
            try:
                validateUnifiedConfig([])
            except Exception as exc:
                raise InvalidUnifiedSchema(str(exc)) from None
        except InvalidUnifiedSchema as exc2:
            errorMsg = str(exc2)
        self.assertTrue("Unified schema error." in errorMsg)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_Monitoring_t_

Unit tests for the WMBS Monitoring DAO objects.
"""

import os
import unittest
import threading

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMBS.Subscription import Subscription

from WMCore.JobStateMachine.Transitions import Transitions
from WMCore.WMBS.CreateWMBSBase import CreateWMBSBase

from WMCore.DAOFactory import DAOFactory
from WMCore.WMFactory import WMFactory
from WMQuality.TestInit import TestInit
from WMCore.Services.UUIDLib import makeUUID

class MonitoringTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        WMBS tables.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = ["WMCore.WMBS"],
                                useDefault = False)

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)
        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """
        self.testInit.clearDatabase()
        return

    def testListJobStates(self):
        """
        _testListJobStates_

        Verify that the ListJobStates DAO works correctly.
        """
        listJobStates = self.daoFactory(classname = "Monitoring.ListJobStates")
        jobStates = listJobStates.execute()

        transitionStates = Transitions().states()
        assert len(jobStates) == len(transitionStates), \
               "Error: Number of states don't match."

        for jobState in jobStates:
            assert jobState in transitionStates, \
                   "Error: Missing job state %s" % jobState

        return

    def testListSubTypes(self):
        """
        _testSubTypes_

        Verify that the ListSubTypes DAO works correctly.
        """
        listSubTypes = self.daoFactory(classname = "Monitoring.ListSubTypes")
        subTypes = listSubTypes.execute()

        schemaTypes = [x[0] for x in CreateWMBSBase().subTypes]
        assert len(subTypes) == len(schemaTypes), \
               "Error: Number of subscription types don't match."
        for subType in subTypes:
            assert subType in schemaTypes, \
                   "Error: Missing subscription type: %s" % subType

        return

    def testListRunningJobs(self):
        """
        _testListRunningJobs_

        Test the ListRunningJobs DAO.
        """
        testWorkflow = Workflow(spec = makeUUID(), owner = "Steve",
                                name = makeUUID(), task="Test")
        testWorkflow.create()

        testFileset = Fileset(name = "TestFileset")
        testFileset.create()

        testSubscription = Subscription(fileset = testFileset,
                                        workflow = testWorkflow,
                                        type = "Processing")
        testSubscription.create()

        testJobGroup = JobGroup(subscription = testSubscription)
        testJobGroup.create()

        testJobA = Job(name = makeUUID(), files = [])
        testJobA["couch_record"] = makeUUID()
        testJobA.create(group = testJobGroup)
        testJobA["state"] = "executing"

        testJobB = Job(name = makeUUID(), files = [])
        testJobB["couch_record"] = makeUUID()
        testJobB.create(group = testJobGroup)
        testJobB["state"] = "complete"

        testJobC = Job(name = makeUUID(), files = [])
        testJobC["couch_record"] = makeUUID()
        testJobC.create(group = testJobGroup)
        testJobC["state"] = "new"

        changeStateAction = self.daoFactory(classname = "Jobs.ChangeState")
        changeStateAction.execute(jobs = [testJobA, testJobB, testJobC])

        runningJobsAction = self.daoFactory(classname = "Monitoring.ListRunningJobs")
        runningJobs = runningJobsAction.execute()

        assert len(runningJobs) == 2, \
               "Error: Wrong number of running jobs returned."

        for runningJob in runningJobs:
            if runningJob["job_name"] == testJobA["name"]:
                assert runningJob["state"] == testJobA["state"], \
                       "Error: Running job has wrong state."
                assert runningJob["couch_record"] == testJobA["couch_record"], \
                       "Error: Running job has wrong couch record."
            else:
                assert runningJob["job_name"] == testJobC["name"], \
                       "Error: Running job has wrong name."
                assert runningJob["state"] == testJobC["state"], \
                       "Error: Running job has wrong state."
                assert runningJob["couch_record"] == testJobC["couch_record"], \
                       "Error: Running job has wrong couch record."

        return

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_DBSError_t_

Unit test for the DBSError class.
"""

import unittest

# from nose.plugins.attrib import attr

from dbs.apis.dbsClient import DbsApi
from RestClient.ErrorHandling.RestClientExceptions import HTTPError
from WMCore.Services.DBS.DBSErrors import DBSError


class DBSErrorsTest(unittest.TestCase):
    """
    DNSErrorsTest class
    """

    def setUp(self):
        """
        _setUp_
        Initialize unit test data
        """

        url = "https://cmsweb-testbed.cern.ch:8443/dbs/int/global/DBSWriter"
        self.api = DbsApi(url=url, useGzip=True, aggregate=False)

        # provide mailformed data to insertBulkBlock API
        files = [
                 {"check_sum": "1504266448",
                  "file_lumi_list": [{"lumi_section_num": 27414, "run_num": 1}],
                  "adler32": "NOTSET",
                  "event_count": 1619,
                  "file_type": "EDM",
                  "last_modified_by": "Test",
                  "logical_file_name": "/store/data/a/b/A/a/1/abcd9.root",
                  "file_size": 123,
                  "last_modification_date": 1279912089,
                  "auto_cross_section": 0.0}]
        self.data = {'files': files}

    def testErrors(self):
        """
        testErrors tests DBSError class
        """
        data = """[{"error": {"code":123, "message": "message", "reason":"reason"}, "http": {"code": 500}}]"""
        dbsError = DBSError(data)
        httpCode = dbsError.getHttpCode()
        srvCode = dbsError.getServerCode()
        msg = dbsError.getMessage()
        reason = dbsError.getReason()
        self.assertTrue(srvCode == 123)
        self.assertTrue(httpCode == 500)
        self.assertTrue(msg == 'message')
        self.assertTrue(reason == 'reason')

    def testDBSError(self):
        """
        testDBSError tests DBSError class
        """
        data = """[{"error":{"reason":"invalid parameter(s)","message":"parameter 'bla' is not accepted by 'datatiers' API","function":"dbs.parameters.CheckQueryParameters","code":118}, "http":{"method":"GET","code":400,"timestamp":"2022-11-10 13:51:27.924928 -0500 EST m=+40.711338056","path":"/dbs2go/datatiers?bla=1","user_agent":"curl/7.86.0","x_forwarded_host":"","x_forwarded_for":"","remote_addr":"127.0.0.1:56191"},"exception":400,"type":"HTTPError","message":"DBSError Code:118 Description:DBS invalid parameter for the DBS API Function:dbs.parameters.CheckQueryParameters Message:parameter 'bla' is not accepted by 'datatiers' API Error: invalid parameter(s)"}]"""
        dbsError = DBSError(data)
        srvCode = dbsError.getServerCode()
        httpCode = dbsError.getHttpCode()
        msg = dbsError.getMessage()
        reason = dbsError.getReason()
        self.assertTrue(srvCode == 118)
        self.assertTrue(httpCode == 400)
        self.assertTrue(msg == "parameter 'bla' is not accepted by 'datatiers' API")
        self.assertTrue(reason == "invalid parameter(s)")

    def testDBSErrorViaDBSClient(self):
        """
        testDBSErrorViaDBSClient tests DBSError class
        """
        # results should contain an error
        # here expected message and reason taken from actual DBS error
        expectedReason = 'DBSError Code:110 Description:DBS DB insert record error Function:dbs.GetRecID Message: Error: nested DBSError Code:113 Description:DBS validation error, e.g. input parameter does not match lexicon rules Function:dbs.primarydstypes.Insert Message: Error: nested DBSError Code:115 Description:DBS decode record failure, e.g. mailformed JSON Function:dbs.DecodeValidatorError Message: Error: validation error'
        expectedMessage = "4106c5ea3bcefee148a7ec69592baa5ec1f1849a240e3af15de850f8e6e5fd81 unable to find primary_ds_type_id for primary ds type=''"
        try:
            self.api.insertBulkBlock(self.data)
        except HTTPError as exp:
            #print("### HTTPError exception: %s" % exp)
            #print("### HTTPError exception representation: %s" % repr(exp))
            dbsError = DBSError(exp)
            srvCode = dbsError.getServerCode()
            httpCode = dbsError.getHttpCode()
            msg = dbsError.getMessage()
            reason = dbsError.getReason()
            self.assertTrue(srvCode == 109)
            self.assertTrue(httpCode == 400)
            self.assertTrue(msg == expectedMessage)
            self.assertTrue(reason == expectedReason)

    def testDBSErrorException(self):
        """
        testDBSErrorException tests DBSError class
        """
        with self.assertRaises(HTTPError):
            self.api.insertBulkBlock(self.data)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import os
import threading
import unittest

from WMCore_t.WMSpec_t.TestSpec import createTestWorkload
from WMComponent.WorkflowUpdater.WorkflowUpdaterPoller import WorkflowUpdaterPoller
from WMCore.Agent.HeartbeatAPI import HeartbeatAPI
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMSpec.Makers.TaskMaker import TaskMaker
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit


class WorkflowUpdaterTest(EmulatedUnitTestCase):
    """
    Test case for the WorkflowUpdater
    """

    def setUp(self):
        """
        Setup the database and logging connection. Try to create all of the
        WMBS tables.
        """
        super(WorkflowUpdaterTest, self).setUp()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase=True)

        self.testInit.setSchema(customModules=['WMCore.WMBS', 'WMCore.Agent.Database'],
                                useDefault=False)

        self.configFile = EmulatorSetup.setupWMAgentConfig()

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.listActiveWflows = self.daoFactory(classname="Workflow.GetUnfinishedWorkflows")

        self.testDir = self.testInit.generateWorkDir()
        self.cwd = os.getcwd()

        # Set heartbeat
        self.componentName = 'WorkflowUpdater'
        self.heartbeatAPI = HeartbeatAPI(self.componentName)
        self.heartbeatAPI.registerComponent()

        # don't execute tearDown because it hangs
        # self._teardown = False

        return

    def tearDown(self):
        """
        Drop all of the non-sql and sql databases
        """
        self.testInit.clearDatabase(modules=['WMCore.WMBS', 'WMCore.Agent.Database'])
        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        EmulatorSetup.deleteConfig(self.configFile)

    def getConfig(self):
        """
        Creates a common config.
        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        # First the general stuff
        config.section_("General")
        config.General.workDir = os.getenv("TESTDIR", os.getcwd())
        config.section_("Agent")
        config.Agent.componentName = self.componentName

        # Now the CoreDatabase information
        # This should be the dialect, dburl, etc
        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        config.component_("WorkflowUpdater")
        config.WorkflowUpdater.namespace = 'WMComponent.WorkflowUpdater.WorkflowUpdater'
        config.WorkflowUpdater.logLevel = 'INFO'
        config.WorkflowUpdater.componentDir = self.testDir
        config.WorkflowUpdater.rucioAccount = "wma_test"
        config.WorkflowUpdater.rucioUrl = "http://cms-rucio-int.cern.ch"
        config.WorkflowUpdater.rucioAuthUrl = "https://cms-rucio-auth-int.cern.ch"
        config.WorkflowUpdater.msPileupUrl = "https://cmsweb-testbed.cern.ch/ms-pileup/data/pileup"
        config.WorkflowUpdater.sandboxDir = "/tmp"

        return config

    def createWorkload(self, workloadName='Test'):
        """
        Creates a test workload for us to run on.
        """
        workload = createTestWorkload(workloadName)
        rereco = workload.getTask("ReReco")
        seederDict = {"generator.initialSeed": 1001, "evtgenproducer.initialSeed": 1001}
        rereco.addGenerator("PresetSeeder", **seederDict)

        taskMaker = TaskMaker(workload, os.path.join(self.testDir, 'workloadTest'))
        # taskMaker.skipSubscription = False
        taskMaker.processWorkload()

        return workload

    def stuffWMBS(self, workflowURL, name):
        """
        Insert some dummy jobs, jobgroups, filesets, files and subscriptions
        into WMBS to test job creation.  Three completed job groups each
        containing several files are injected.  Another incomplete job group is
        also injected.  Also files are added to the "Mergeable" subscription as
        well as to the output fileset for their jobgroups.
        """
        testSite = "T2_CH_CERN"
        locationAction = self.daoFactory(classname="Locations.New")
        locationAction.execute(siteName=testSite, pnn=testSite)

        mergeFileset = Fileset(name="mergeFileset")
        mergeFileset.create()
        bogusFileset = Fileset(name="bogusFileset")
        bogusFileset.create()

        mergeWorkflow = Workflow(spec=workflowURL, name=name, task="/TestWorkload/ReReco")
        mergeWorkflow.create()

        mergeSubscription = Subscription(fileset=mergeFileset,
                                         workflow=mergeWorkflow,
                                         split_algo="ParentlessMergeBySize")
        mergeSubscription.create()
        dummySubscription = Subscription(fileset=bogusFileset,
                                         workflow=mergeWorkflow,
                                         split_algo="ParentlessMergeBySize")

        fileObjList = []
        for idx, fName in enumerate(["file1", "file2", "file3", "file4"]):
            testFile = File(lfn=fName, size=1024, events=1024, first_event=1024 * idx, locations={testSite})
            testFile.addRun(Run(1, *[45]))
            testFile.create()
            fileObjList.append(testFile)
        for idx, fName in enumerate(["fileA", "fileB", "fileC"]):
            testFile = File(lfn=fName, size=1024, events=1024, first_event=1024 * idx, locations={testSite})
            testFile.addRun(Run(1, *[46]))
            testFile.create()
            fileObjList.append(testFile)
        for idx, fName in enumerate(["fileI", "fileII", "fileIII", "fileIV"]):
            testFile = File(lfn=fName, size=1024, events=1024, first_event=1024 * idx, locations={testSite})
            testFile.addRun(Run(2, *[46]))
            testFile.create()
            fileObjList.append(testFile)

        for fileObj in fileObjList:
            mergeFileset.addFile(fileObj)
            bogusFileset.addFile(fileObj)

        mergeFileset.commit()
        bogusFileset.commit()

        return

    def testVerySimpleTest(self):
        """
        Inject a workflow into the database and run WorkflowUpdater algorithm
        """
        # populate the database
        workloadName = 'TestWorkload'
        _dummyWorkload = self.createWorkload(workloadName=workloadName)
        workloadPath = os.path.join(self.testDir, 'workloadTest', 'TestWorkload', 'WMSandbox', 'WMWorkload.pkl')
        self.stuffWMBS(workflowURL=workloadPath, name=makeUUID())

        myThread = threading.currentThread()

        config = self.getConfig()
        testWflowUpdater = WorkflowUpdaterPoller(config=config)
        testWflowUpdater.algorithm()
        # self.assertFalse(True)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=
"""
File       : SummaryDB.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Set of modules/functions to update WMAgent SummaryDB.
"""

from future.utils import viewitems, viewvalues

# system modules
import logging
from pprint import pformat

from WMCore.Database.CMSCouch import CouchNotFoundError


def fwjr_parser(doc):
    """
    Parse FWJR document and yield the following structure:
    {"id": "requestName",
        "tasks": {
            "taskName1": {
                "sites": {
                    "siteName1":  {
                        "wrappedTotalJobTime": 1612,
                        "cmsRunCPUPerformance": {"totalJobCPU": 20, "totalJobTime": 42, "totalEventCPU": 4},
                        "inputEvents": 0,
                        "dataset": {}
                    }
                    "siteName2":  {
                        "wrappedTotalJobTime": 1612,
                        "cmsRunCPUPerformance": {"totalJobCPU": 20, "totalJobTime": 42, "totalEventCPU": 4},
                        "inputEvents": 0,
                        "dataset": {
                             "/a/b/c": {"size": 50, "events": 10, "totalLumis": 100},
                             "/a1/b1/c1": {"size": 50, "events": 10, "totalLumis": 100}
                        },
                    }
                },
            }
            "taskName2" : {}
        }
    }
    """
    # flag for the data collection, in some case fwjr doesn't contains the data needed
    # In that case it shouldn't crash but move on.
    dataCollectedFlag = False

    if 'fwjr' in doc:
        fwjr = doc['fwjr']
        if not isinstance(fwjr, dict):
            fwjr = fwjr.__to_json__(None)
    else:
        raise Exception('Document does not contain FWJR part')

    if 'task' not in fwjr:
        return dataCollectedFlag

    task_name = fwjr['task']
    try:
        req_name = task_name.split('/')[1]
    except:
        raise Exception('Cannot get request name from task_name "%s"' % task_name)

    # if one of the step has error except logArchive step don't collect the data
    for stepName in fwjr['steps']:
        if not stepName.startswith('logArch') and fwjr['steps'][stepName]['status'] != 0:
            return dataCollectedFlag

    sdict = {}  # all sites summary
    steps = fwjr['steps']
    wrappedTotalJobTime = 0
    pdict = dict(totalJobCPU=0, totalJobTime=0, totalEventCPU=0)
    ddict = {}
    for key, val in viewitems(steps):
        if val['stop'] is not None and val['start'] is not None:
            wrappedTotalJobTime += val['stop'] - val['start']
        site_name = val['site']
        site_summary = dict(wrappedTotalJobTime=wrappedTotalJobTime,
                            inputEvents=0, dataset=ddict,
                            cmsRunCPUPerformance=pdict)
        if key.startswith('cmsRun'):
            perf = val['performance']
            perf.setdefault('cpu', {})
            pdict['totalJobCPU'] += float(perf['cpu'].get('TotalJobCPU', 0))
            pdict['totalJobTime'] += float(perf['cpu'].get('TotalJobTime', 0))
            if 'TotalEventCPU' in perf['cpu']:
                pdict['totalEventCPU'] += float(perf['cpu']['TotalEventCPU'])
            else:
                pdict['totalEventCPU'] += float(perf['cpu'].get('TotalLoopCPU', 0))

            odict = val['output']
            for kkk, vvv in viewitems(odict):
                for row in vvv:
                    if row.get('merged', False):
                        prim = row['dataset']['primaryDataset']
                        proc = row['dataset']['processedDataset']
                        tier = row['dataset']['dataTier']
                        dataset = '/%s/%s/%s' % (prim, proc, tier)
                        totalLumis = sum([len(r) for r in viewvalues(row['runs'])])
                        dataset_summary = \
                            dict(size=row['size'], events=row['events'], totalLumis=totalLumis)
                        ddict[dataset] = dataset_summary
            if ddict:  # if we got dataset summary
                site_summary.update({'dataset': ddict})

            idict = val.get('input', {})
            if idict:
                source = idict.get('source', {})
                if isinstance(source, list):
                    for item in source:
                        site_summary['inputEvents'] += item.get('events', 0)
                else:
                    site_summary['inputEvents'] += source.get('events', 0)
        sdict[site_name] = site_summary
        dataCollectedFlag = True

    if dataCollectedFlag:
        # prepare final data structure
        sum_doc = dict(_id=req_name, tasks={task_name: dict(sites=sdict)})
        return sum_doc
    else:
        return dataCollectedFlag


def merge_docs(doc1, doc2):
    "Merge two summary documents use dict in place strategy"
    for key in ['_id', '_rev']:
        if key in doc1:
            del doc1[key]
        if key in doc2:
            del doc2[key]
    for key, val in viewitems(doc1):
        if key not in doc2:
            return False
        if isinstance(val, dict):
            if not merge_docs(doc1[key], doc2[key]):
                return False
        else:
            doc1[key] += doc2[key]
    return True


def update_tasks(old_tasks, new_tasks):
    "Update tasks dictionaries"
    for task, tval in viewitems(new_tasks):
        if task in old_tasks:
            for site, sval in viewitems(tval['sites']):
                if site in old_tasks[task]['sites']:
                    old_sdict = old_tasks[task]['sites'][site]
                    new_sdict = sval
                    status = merge_docs(old_sdict, new_sdict)
                    if not status:
                        logging.error("Error merge_docs:\n%s\n%s\n", pformat(old_sdict),
                                      pformat(new_sdict))
                else:
                    old_tasks[task]['sites'].update({site: sval})
        else:
            old_tasks.update({task: tval})
    return old_tasks


def updateSummaryDB(sumdb, document):
    "Update summary DB with given document"
    # parse input doc and create summary doc
    sum_doc = fwjr_parser(document)
    # check if DB has FWJR statistics
    if sum_doc is False:
        return False

    try:
        doc = sumdb.document(sum_doc['_id'])
        old_tasks = doc['tasks']
    except CouchNotFoundError:
        old_tasks = {}

    tasks = update_tasks(old_tasks, sum_doc['tasks'])

    try:
        combinedDoc = {'_id': sum_doc['_id'], 'tasks': tasks}
        resp = sumdb.updateDocument(sum_doc['_id'], "SummaryStats",
                                    "genericUpdate",
                                    fields=combinedDoc,
                                    useBody=True)
        # TODO: check response whether update successfull or not
        return True
    except Exception:
        logging.exception("Error fetching summary doc %s:", sum_doc['_id'])
        return False



#!/usr/bin/env python
"""
_GetCache_

MySQL implementation of Jobs.GetState
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

import logging

class SetCache(DBFormatter):
    """
    _GetState_

    Given a job ID, get the state of a current job.
    """
    sql = """UPDATE wmbs_job
             SET cache_dir = :cacheDir
             WHERE id = :jobid"""

    def execute(self, id = None, cacheDir = None, conn = None, transaction = False, jobDictList = None):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        Sending a value for jobDictList triggers bulk mode
        """
        #jobDictList != None triggers bulk mode
        #jobDictList should be a list of dictionaries of the form {'id': id,'cacheDir': cacheDir}

        if jobDictList:
            binds = jobDictList
        elif id and cacheDir:
            binds = {"jobid": id, "cacheDir": cacheDir}
        else:
            logging.error("Jobs.SetCache not sent values to set!")
            return

        result = self.dbi.processData(self.sql, binds, conn = conn, transaction = transaction)

        return



#!/usr/bin/env python
"""
_BasicProductionWorkload_

Sample/Tester for a simple production workflow and associated merge.

Production task produces a single output dataset,
Merge task is used to merge that dataset

"""
from WMCore.WMSpec.WMWorkload import newWorkload
from WMCore.WMSpec.WMStep import makeWMStep
from WMCore.WMSpec.Steps.StepFactory import getStepTypeHelper

#  //
# // Set up the basic workload task and step structure
#//
workload = newWorkload("BasicProduction")
workload.setOwner("WMTest")
workload.setStartPolicy('MonteCarlo', **{'SliceType': 'NumberOfEvents', 'SliceSize': 100})
workload.setEndPolicy('SingleShot')
workload.setSiteWhitelist(["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"])
#  //
# // set up the production task
#//
production = workload.newTask("Production")
production.addProduction(totalevents = 1000)
#WARNING: this is arbitrary task type (wmbs schema only supprot "Processing", "Merge", "Harvest") - maybe add "MCProduction"
production.setTaskType("Merge")
production.setSplittingAlgorithm("EventBased", events_per_job = 100)
production.setSiteWhitelist(["T2_XX_SiteA", "T2_XX_SiteB", "T2_XX_SiteC"])
prodCmssw = production.makeStep("cmsRun1")
prodCmssw.setStepType("CMSSW")
prodStageOut = prodCmssw.addStep("stageOut1")
prodStageOut.setStepType("StageOut")
production.applyTemplates()

#  //
# // set up the merge task
#//
merge = production.addTask("Merge")
merge.setTaskType("Merge")
merge.setSplittingAlgorithm("MergeBySize", merge_size = 20000000)
mergeCmssw = merge.makeStep("cmsRun1")
mergeCmssw.setStepType("CMSSW")
mergeStageOut = mergeCmssw.addStep("stageOut1")
mergeStageOut.setStepType("StageOut")
merge.applyTemplates()


#  //
# // populate the details of the production tasks
#//
#  //
# // production cmssw step
#//
#
# TODO: Anywhere helper.data is accessed means we need a method added to the
# type based helper class to provide a clear API.
prodCmsswHelper = prodCmssw.getTypeHelper()

prodCmsswHelper.data.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
prodCmsswHelper.data.application.setup.softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"

prodCmsswHelper.data.application.configuration.configCacheUrl = "http://whatever"

prodCmsswHelper.addOutputModule("writeData", primaryDataset = "Primary",
                                processedDataset = "Processed",
                                dataTier = "TIER")
#print prodCmsswHelper.data



#  //
# // production stage out step
#//
prodStageOutHelper = prodStageOut.getTypeHelper()


#  //
# // merge cmssw step
#//
# point it at the input step from the previous task
merge.setInputReference(prodCmssw, outputModule = "writeData")


#  //
# // populate the details of the merge tasks
#//



# print workload.data



#!/usr/bin/env python

"""
Request Manager service (ReqMgr) test and client script.

The script shall have no WMCore libraries dependency.

Command line interface: --help

There are mandatory command line arguments (e.g. URL of the Request Manager)

Production ConfigCache: https://cmsweb.cern.ch/couchdb/reqmgr_config_cache/

Note: tests for checking data directly in CouchDB in ReqMgr1 test script:
    WMCore/test/data/ReqMgr/reqmgr.py
"""


from builtins import object, str as newstr, bytes as newbytes, next

import json
import logging
import os
import sys
import urllib.parse
from argparse import ArgumentParser
from http.client import HTTPSConnection, HTTPConnection


class RESTClient(object):
    """
    HTTP client
    HTTPS client based on the provided URL (http:// or https://)
    """

    def __init__(self, url, cert=None, key=None, logger=None):
        self.logger = logger
        self.logger.info("RESTClient URL: %s", url)
        if url.startswith("https://"):
            self.logger.debug("Using HTTPS protocol, getting user identity files ...")
            proxy_file = "/tmp/x509up_u%s" % os.getuid()
            if not os.path.exists(proxy_file):
                proxy_file = "UNDEFINED"
            cert_file = cert or os.getenv("X509_USER_CERT",
                                          os.getenv("X509_USER_PROXY", proxy_file))
            key_file = key or os.getenv("X509_USER_KEY",
                                        os.getenv("X509_USER_PROXY", proxy_file))
            self.logger.info("Identity files:\n\tcert file: '%s'\n\tkey file:  '%s' ",
                             cert_file, key_file)
            url = url.replace("https://", '')
            self.logger.debug("Creating connection HTTPS ...")
            self.conn = HTTPSConnection(url, key_file=key_file, cert_file=cert_file)
        if url.startswith("http://"):
            self.logger.info("Using HTTP protocol, creating HTTP connection ...")
            url = url.replace("http://", '')
            self.conn = HTTPConnection(url)

    def http_request(self, verb, uri, data=None, headers=None):
        self.logger.debug("Request: %s %s %s ..." % (verb, uri, data))
        self.conn.request(verb, uri, body=data, headers=headers or self.headers)
        resp = self.conn.getresponse()
        data = resp.read()
        self.logger.debug("Status: %s", resp.status)
        self.logger.debug("Reason: %s", resp.reason)
        if resp.status != 200:
            if hasattr(resp.msg, "x-error-detail"):
                self.logger.warning("Message: %s", resp.msg["x-error-detail"])
        return resp.status, data


class ReqMgrClient(RESTClient):
    """
    Client REST interface to Request Manager service (ReqMgr).

    Actions: all_tests

    For reference:
        jsonArgs = json.dumps(requestArgs["createRequest"])
        status, data = self.http_request("PUT", "/reqmgr/reqMgr/request", data=jsonArgs)
        data = json.loads(data)
        requestName = data.values()[0]["request"]

        params = {"requestName": requestName,
                  "status": "assignment-approved"}
        encodedParams = urllib.urlencode(params)
        logging.info("Approving request '%s' ..." % requestName)
        status, data = self.http_request("PUT", "/reqmgr/reqMgr/request",
                                        data=encodedParams, headers=self.textHeaders)

    """

    def __init__(self, url, config, logger=None):
        self.logger = logging.getLogger() if not logger else logger
        self.logger.info("ReqMgr url: %s", url)
        # ReqMgr based on WMCore.REST API requires accept types defined
        self.headersUrl = {"Content-type": "application/x-www-form-urlencoded",
                           "Accept": "application/json"}
        self.headersBody = {"Content-type": "application/json",
                            "Accept": "application/json"}
        self.headers = self.headersUrl
        self.urn_prefix = "/reqmgr2/data"
        RESTClient.__init__(self, url, cert=config.cert, key=config.key, logger=logger)

    def _caller_checker(self, urn, verb, input_data=None, exp_data=None):
        urn = self.urn_prefix + urn
        self.logger.info("Call %s %s %s", urn, verb, input_data)
        status, data = self.http_request(verb, urn, data=input_data)
        if status != 200:
            self.logger.error("HTTP request failed with status: %s, data: %s", status, data)
            return
        data = json.loads(data)["result"]
        if exp_data:
            assert status == 200, "Call status is: %s" % status
            assert data[0] == exp_data, "'%s' != '%s' mismatch." % (data[0], exp_data)
        else:
            assert status == 200, "Call status is: %s" % status
            self.logger.info("status: %s\n%s", status, data)
        return data

    def delete_requests(self, config):
        urn = self.urn_prefix + "/request"
        for request_name in config.request_names:
            self.logger.info("Deleting '%s' request ...", request_name)
            args = urllib.parse.urlencode({"request_name": request_name})
            status, data = self.http_request("DELETE", urn, data=args)
            if status != 200:
                self.logger.error("Failed to delete request with status: %s, data: %s", status, data)
                sys.exit(1)
            self.logger.info("Done.")

    def create_request(self, config):
        """
        config.request_args - arguments for both creation and assignment

        """
        self.logger.info("Injecting request args:\n%s ...", config.request_args["createRequest"])
        json_args = json.dumps(config.request_args["createRequest"])
        urn = self.urn_prefix + "/request"
        status, data = self.http_request("POST", urn, data=json_args,
                                         headers=self.headersBody)
        if status > 216:
            self.logger.error("Failed to create request with status: %s, data: %s", status, data)
            sys.exit(1)
        data = json.loads(data)
        self.logger.info(data)
        request_name = data["result"][0]["request"]
        self.approve_request(request_name)
        self.logger.info("Create request '%s' succeeded.", request_name)

        config.request_names = request_name

        return request_name

    def approve_request(self, request_name):
        """
        Set request status assignment-approved of the requestName request.
        Once ReqMgr provides proper API for status settings, esp. for assignment,
        a single method setStates shall handle all request status changes.

        """
        self.logger.info("Approving request '%s' ...", request_name)

        json_args = json.dumps({"RequestStatus": "assignment-approved"})
        urn = self.urn_prefix + "/request/%s" % request_name
        status, data = self.http_request("PUT", urn, data=json_args,
                                         headers=self.headersBody)

        if status != 200:
            self.logger.error("Failed to approve request with status: %s, data: %s", status, data)
            sys.exit(1)
        self.logger.info("Approve succeeded.")

    def assign_request(self, config):
        """
        config.request_args - arguments for both creation and assignment

        """
        assign_args = config.request_args["assignRequest"]
        assign_args["RequestStatus"] = "assigned"
        json_args = json.dumps(assign_args)
        if isinstance(config.request_names, (newstr, newbytes)):
            config.request_names = [config.request_names]
        for request_name in config.request_names:
            self.logger.info("Assigning %s with request args: %s ...",
                             request_name, config.request_args["assignRequest"])
            urn = self.urn_prefix + "/request/%s" % request_name
            status, data = self.http_request("PUT", urn, data=json_args,
                                             headers=self.headersBody)
            if status > 216:
                self.logger.error("Failed to assign request with status: %s, data: %s", status, data)
                sys.exit(1)
            data = json.loads(data)
            self.logger.info(data)
            self.logger.info("Assign succeeded.")

    def query_requests(self, config, to_query=None):
        """
        If to_query and config.request_names are not specified, then
        all requests in the system are queried.
        toQuery - particular request name to query.
        config.request_names - list of requests to query.

        Returns a list of requests in either case.

        """
        if to_query:
            requests_to_query = [to_query]
        else:
            requests_to_query = config.request_names

        requests_data = []
        if requests_to_query:
            for request_name in requests_to_query:
                self.logger.info("Querying '%s' request ...", request_name)
                urn = self.urn_prefix + "/request?name=%s" % request_name
                status, data = self.http_request("GET", urn)
                if status != 200:
                    self.logger.error("Failed to get request with status: %s, data: %s", status, data)
                    sys.exit(1)
                request = json.loads(data)["result"][0]
                for k, v in sorted(request.items()):
                    self.logger.info("\t%s: %s", k, v)
                requests_data.append(request)
            # returns data on requests in the same order as in the config.request_names
            return requests_data

    def all_tests(self, config):
        self._caller_checker("/hello", "GET", exp_data="Hello world")
        self._caller_checker("/hello?name=John", "GET", exp_data="Hello John")
        self._caller_checker("/about", "GET")
        self._caller_checker("/info", "GET")
        group = "mygroup"
        args = urllib.parse.urlencode({"group_name": group})
        self._caller_checker("/group", "PUT", input_data=args)
        data = self._caller_checker("/group", "GET")
        assert group in data, "%s should be in %s" % (group, data)
        self._caller_checker("/group", "DELETE", input_data=args)
        data = self._caller_checker("/group", "GET")
        assert group not in data, "%s should be deleted from %s" % (group, data)
        team = "myteam"
        args = urllib.parse.urlencode({"team_name": team})
        self._caller_checker("/team", "PUT", input_data=args)
        data = self._caller_checker("/team", "GET")
        assert team in data, "%s should be in %s" % (team, data)
        self._caller_checker("/team", "DELETE", input_data=args)
        data = self._caller_checker("/team", "GET")
        assert team not in data, "%s should be deleted from %s" % (team, data)

        data = self._caller_checker("/software", "GET")

        data = self._caller_checker("/status", "GET")
        assert len(data) > 0, "%s should be non-empty list." % data
        # test some request status
        status = ["assigned", "assignment-approved", "failed", "new"]
        for s in status: assert s in data, "%s is not in %s" % (s, data)
        data2 = self._caller_checker("/status?transition=false", "GET")
        assert data == data2, "%s != %s" % (data, data2)
        # returns also all allowed transitions
        data = self._caller_checker("/status?transition=true", "GET")
        for status_def in data:
            status = next(iter(status_def))
            trans = status_def[status]
            assert status in data2, "%s is not in %s" % (status, data2)
            assert isinstance(trans, list), "transition %s should be list" % trans

        data = self._caller_checker("/type", "GET")
        assert len(data) > 0, "%s should be non-empty list." % data

        # request tests
        new_request_name = self.create_request(config)

        data = self._caller_checker("/request?name=%s" % new_request_name, "GET")
        request = data[0][new_request_name]
        assert request["RequestName"] == new_request_name
        assert request["RequestStatus"] == "new"

        self.logger.info("\nall_tests succeeded.")

    def __del__(self):
        self.conn.close()
        del self.conn


# ---------------------------------------------------------------------------


def process_cli_args():
    def err_exit(msg, parser):
        print('\n')
        parser.print_help()
        print("\n\n%s" % msg)
        sys.exit(1)

    parser = ArgumentParser(usage='%(prog)s [options]', add_help=False)
    actions = define_cli_options(parser)
    # opts - new processed options
    # args - remainder of the input array
    opts = parser.parse_args()
    # check command line arguments validity
    if not opts.reqmgrurl:
        err_exit("Missing mandatory --reqmgrurl.", parser)

    if opts.create_request and not opts.config_file:
        err_exit("When --create_request, --config_file is necessary.", parser)
    if opts.create_request and opts.request_names:
        err_exit("--request_names can't be provided with --create_request", parser)
    if opts.all_tests and not opts.config_file:
        err_exit("When --all_tests, --config_file is necessary", parser)
    if opts.json and not (opts.create_request or opts.assign_request or opts.all_tests):
        err_exit("--json only with --create_request, --assign_request or --all_tests", parser)

    for action in [name for name in actions if getattr(opts, name)]:
        if opts.all_tests and action and action != "all_tests":
            err_exit("Arguments --all_tests and --%s mutually exclusive." % action, parser)

    if opts.request_names:
        # make it a list here
        opts.request_names = opts.request_names.split(',')

    return opts, actions


def define_cli_options(parser):
    actions = []
    # "-h" ------------------------------------------------------------------
    help = "Display this help"
    parser.add_argument("-h", "--help", help=help, action='help')
    # "-c" ------------------------------------------------------------------
    help = ("User cert file (or cert proxy file). "
            "If not defined, tries X509_USER_CERT then X509_USER_PROXY env. "
            "variables. And lastly /tmp/x509up_uUID.")
    parser.add_argument("-c", "--cert", help=help)
    # "-k" ------------------------------------------------------------------
    help = ("User key file (or cert proxy file). "
            "If not defined, tries X509_USER_KEY then X509_USER_PROXY env. "
            "variables. And lastly /tmp/x509up_uUID.")
    parser.add_argument("-k", "--key", help=help)
    # -u --------------------------------------------------------------------
    help = ("Request Manager service address (if not options is supplied, "
            "returns a list of the requests in ReqMgr) "
            "e.g.: https://maxareqmgr01.cern.ch")
    parser.add_argument("-u", "--reqmgrurl", help=help)
    # -f --------------------------------------------------------------------
    help = "Request template with arguments definition, request config file."
    parser.add_argument("-f", "--config_file", help=help)
    # -j --------------------------------------------------------------------
    help = ("JSON string to override values from --config_file. "
            "e.g. --json=\'{\"createRequest\": {\"Requestor\": \"efajardo\"}, "
            "\"assignRequest\": {\"FirstLumi\": 1}}\' "
            "e.g. --json=`\"cat alan.json\"`")
    parser.add_argument("-j", "--json", help=help)
    # -r --------------------------------------------------------------------
    help = ("Request name or list of comma-separated names to perform "
            "actions upon.")
    parser.add_argument("-r", "--request_names", help=help)
    # -v ---------------------------------------------------------------------
    help = "Verbose console output."
    parser.add_argument("-v", "--verbose", action="store_true", help=help)
    # actions definition below ----------------------------------------------
    # -i --------------------------------------------------------------------
    help = ("Action: Create and approve a request. Whichever from the config "
            "file defined arguments can be overridden from "
            "command line and a few have to be so (*-OVERRIDE-ME ending). "
            "Depends on --config_file.")
    action = "create_request"
    actions.append(action)
    parser.add_argument("-i", "--" + action, action="store_true", help=help)
    # -g --------------------------------------------------------------------
    help = ("Action: Assign request(s) specified by --request_names "
            "or a new request when used with --create_request. "
            "Depends on --request_names and --config_file when used without "
            "--create_request")
    action = "assign_request"
    actions.append(action)
    parser.add_argument("-g", "--" + action, action="store_true", help=help)

    # -d --------------------------------------------------------------------
    help = "Action: Delete request(s) specified by --request_names."
    action = "delete_requests"
    actions.append(action)
    parser.add_argument("-d", "--" + action, action="store_true", help=help)
    # -q --------------------------------------------------------------------
    help = "Action: Query request(s) specified by --request_names."
    action = "query_requests"
    actions.append(action)
    parser.add_argument("-q", "--" + action, action="store_true", help=help)
    # -a --------------------------------------------------------------------
    help = ("Action: Perform all operations this script allows. "
            "Check all REST API of the ReqMgr service.")
    action = "all_tests"
    actions.append(action)
    parser.add_argument("-a", "--" + action, action="store_true", help=help)

    return actions


def process_request_args(input_config_file, command_line_json, logger):
    """
    Load request arguments from a file, blend with JSON from command line.

    """
    logger.info("Loading file '%s' ...", input_config_file)
    try:
        request_args = json.load(open(input_config_file, 'r'))
    except IOError as ex:
        logger.fatal("Reading request arguments file '%s' failed, reason: %s.", input_config_file, ex)
        sys.exit(1)
    if command_line_json:
        logger.info("Parsing request arguments on the command line ...")
        cli_json = json.loads(command_line_json)
        # if a key exists in cli_json, update values in the main request_args dict
        for k in request_args:
            if k in cli_json:
                request_args[k].update(cli_json[k])
    else:
        logger.warning("No request arguments to override (--json)? Some values will be wrong.")

    # iterate over all items recursively and warn about those ending with
    # OVERRIDE-ME, hence not overridden
    def check(items):
        for k, v in items:
            if isinstance(v, dict):
                check(v.items())
            if isinstance(v, newstr) and v.endswith("OVERRIDE-ME"):
                logger.warning("Not properly set: %s: %s", k, v)

    check(request_args.items())
    return request_args


def initialization():
    config, actions = process_cli_args()
    if config.verbose:
        logger = loggerSetup(logging.INFO)
    else:
        logger = loggerSetup()
    logger.info("Command line arguments: %s ...", sys.argv)

    reqmgr_client = ReqMgrClient(config.reqmgrurl, config, logger)
    if config.create_request or config.assign_request or config.all_tests:
        # process request arguments and store them
        config.request_args = process_request_args(config.config_file, config.json, logger)
    return reqmgr_client, config, actions, logger


def loggerSetup(logLevel=logging.INFO):
    """
    Return a logger which writes everything to stdout.
    """
    logger = logging.getLogger(__name__)
    outHandler = logging.StreamHandler(sys.stdout)
    outHandler.setFormatter(logging.Formatter("%(asctime)s:%(levelname)s:%(module)s: %(message)s"))
    outHandler.setLevel(logLevel)
    logger.addHandler(outHandler)
    logger.setLevel(logLevel)
    return logger


def main():
    reqmgr_client, config, defined_actions, logger = initialization()
    # definedAction are all actions as defined for CLI
    # there is now gonna be usually 1 action to perform, but could be more
    # filter out those where config.ACTION is None
    # config is all options for this script but also request creation parameters
    actions = [name for name in defined_actions if getattr(config, name)]
    logger.debug("Actions to perform: %s", actions)
    for action in actions:
        logger.info("Performing '%s' ...", action)
        # some methods need to modify config (e.g. add a request name),
        # pass them entire configuration
        reqmgr_client.__getattribute__(action)(config)
    if not actions:
        reqmgr_client.query_requests(config)


if __name__ == "__main__":
    main()



'''
Created on May 22, 2009

@author: meloam
'''
import unittest
import WMCore_t.WMSpec_t.TestWorkloads as TestSpecs
import WMCore.WMSpec.WMWorkload
from WMCore.WMSpec.WMWorkload import getWorkloadFromTask
from WMCore.WMSpec.WMTask import  getTaskFromStep

class WMWorkloadTest(unittest.TestCase):

    def testOneTaskTwoSTep(self):
        workload = TestSpecs.oneTaskTwoStep()
        tasks = workload.listAllTaskNames()
        task = workload.getTask(tasks[0])
        findWorkload = getWorkloadFromTask(task)
        self.assertEqual(id(findWorkload.data), id(workload.data))
        self.assertTrue( isinstance( workload, WMCore.WMSpec.WMWorkload.WMWorkloadHelper ) )
        self.assertTrue( isinstance( task, WMCore.WMSpec.WMTask.WMTaskHelper ) )
        steps = task.listAllStepNames()
        step = task.getStep(steps[0])
        self.assertTrue( isinstance( step, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step2= task.getStep(steps[1])
        self.assertTrue( isinstance( step2, WMCore.WMSpec.WMStep.WMStepHelper ) )
        self.assertEqual( id(getTaskFromStep(step)), id(getTaskFromStep(step2)) )

    def testOneTaskFourStep(self):
        workload = TestSpecs.oneTaskFourStep()
        tasks = workload.listAllTaskNames()
        task = workload.getTask(tasks[0])
        findWorkload = getWorkloadFromTask(task)
        self.assertEqual(id(findWorkload.data), id(workload.data))
        self.assertTrue( isinstance( workload, WMCore.WMSpec.WMWorkload.WMWorkloadHelper ) )
        self.assertTrue( isinstance( task, WMCore.WMSpec.WMTask.WMTaskHelper ) )

        steps = task.listAllStepNames()
        step = task.getStep(steps[0])
        self.assertTrue( isinstance( step, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step2= task.getStep(steps[1])
        self.assertTrue( isinstance( step2, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step3= task.getStep(steps[2])
        self.assertTrue( isinstance( step3, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step4= task.getStep(steps[3])
        self.assertTrue( isinstance( step4, WMCore.WMSpec.WMStep.WMStepHelper ) )
        #print step4
        self.assertEqual( id(getTaskFromStep(step)), id(getTaskFromStep(step2)) )

    def testTwoTaskTree(self):
        workload = TestSpecs.twoTaskTree()
        tasks = workload.listAllTaskNames()
        parenttask = workload.getTask("FirstTask")

        findWorkload = getWorkloadFromTask(parenttask)
        self.assertEqual(id(findWorkload.data), id(workload.data))
        self.assertTrue( isinstance( workload, WMCore.WMSpec.WMWorkload.WMWorkloadHelper ) )
        self.assertTrue( isinstance( parenttask, WMCore.WMSpec.WMTask.WMTaskHelper ) )

        task = workload.getTask("SecondTask")
        #print task.data
        self.assertTrue( isinstance( task, WMCore.WMSpec.WMTask.WMTaskHelper ) )
        steps = task.listAllStepNames()

        # there should be a way to do this with iteration that would be neater
        step = task.getStep(steps[0])
        self.assertTrue( isinstance( step, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step2= task.getStep(steps[1])
        self.assertTrue( isinstance( step2, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step3= task.getStep(steps[2])
        self.assertTrue( isinstance( step3, WMCore.WMSpec.WMStep.WMStepHelper ) )
        step4= task.getStep(steps[3])
        self.assertTrue( isinstance( step4, WMCore.WMSpec.WMStep.WMStepHelper ) )
        self.assertEqual( id(getTaskFromStep(step)), id(getTaskFromStep(step2)) )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_BulkNewReturn_

MySQL implementation of Fileset.BulkNewReturn
"""

__all__ = []



import time

from WMCore.Database.DBFormatter import DBFormatter

class BulkNewReturn(DBFormatter):
    """
    Creates filesets in bulk and returns their IDs

    """
    sql = """INSERT INTO wmbs_fileset (name, last_update, open)
               VALUES (:NAME, :LAST_UPDATE, :OPEN)"""

    returnSQL = """SELECT ID FROM wmbs_fileset
                   WHERE name = :NAME
                   AND last_update = :LAST_UPDATE
                   AND open = :OPEN"""

    def getBinds(self, name = None, open = False):
        bindVars = {}
        bindVars["NAME"] = name
        bindVars["OPEN"] = int(open)
        bindVars["LAST_UPDATE"] = int(time.time())
        return bindVars

    def execute(self, nameList = None, open = False, conn = None,
                transaction = False):
        """
        This can take a bulk argument list that should only be the names

        """
        binds = []
        for name in nameList:
            binds.append(self.getBinds(name = name, open = open))

        self.dbi.processData(self.sql, binds,
                             conn = conn, transaction = transaction)

        result = self.dbi.processData(self.returnSQL, binds,
                                      conn = conn, transaction = transaction)

        res = self.format(result)

        # This should be a list
        final = []
        for entry in res:
            final.append(entry[0])

        return final



from __future__ import print_function
import logging
logging.basicConfig(level = logging.DEBUG)
import unittest
from mock import mock
from WMCore.Storage.Plugins.LCGImpl import LCGImpl as ourPlugin
import WMCore.Storage.Plugins.LCGImpl
moduleWeAreTesting = WMCore.Storage.Plugins.LCGImpl

from WMCore.Storage.StageOutError import StageOutError, StageOutFailure

from nose.plugins.attrib import attr

def getPluginObject(mock_os, mock_Popen, localSize, remoteSize):

    # Mock file size of local file
    mock_os.path.getsize.return_value = localSize

    process_mock = mock.Mock()
    # Mock remote file size
    attrs = {'communicate.return_value': ('0 0 0 0 %s' % remoteSize, 'error')}
    process_mock.configure_mock(**attrs)
    mock_Popen.return_value = process_mock

    # Actually run the test
    ourPlugin.runCommandFailOnNonZero = mock.Mock()
    ourPlugin.runCommandFailOnNonZero.return_value = (0,  None)
    testObject = ourPlugin()
    return testObject

class SRMV2ImplTest(unittest.TestCase):

    @attr("integration")
    @mock.patch('WMCore.Storage.Plugins.LCGImpl.subprocess.Popen')
    @mock.patch('WMCore.Storage.Plugins.LCGImpl.os')
    def testFailSrmCopy(self, mock_os, mock_Popen):
        
        mock_os.path.getsize.return_value = 9001

        # copy a file and have it fail
        testObject = ourPlugin()

        # copy normally and have it work
        self.assertRaises(StageOutError, testObject.doTransfer,'file:///store/NONEXISTANTSOURCE',
                              'srm://nonexistant.com/blah/?SFN=/store/NONEXISTANTTARGET',
                              True,
                              None,
                              None,
                              None,
                              None,
                              None)


    @attr("integration")
    @mock.patch('WMCore.Storage.Plugins.LCGImpl.subprocess.Popen')
    @mock.patch('WMCore.Storage.Plugins.LCGImpl.os')
    def testFailOnFileSize(self, mock_os, mock_Popen, localSize=9001, remoteSize=9002):

        # copy a file and have it fail file size check
        testObject = getPluginObject(mock_os, mock_Popen, localSize, remoteSize)

        # Do transfer and fail local<->remote file size check
        self.assertRaises(StageOutFailure, testObject.doTransfer,'file:///store/NONEXISTANTSOURCE',
                              'srm://nonexistant.com/blah/?SFN=/store/NONEXISTANTTARGET',
                              True,
                              None,
                              None,
                              None,
                              None,
                              None)

    @attr("integration")
    @mock.patch('WMCore.Storage.Plugins.LCGImpl.subprocess.Popen')
    @mock.patch('WMCore.Storage.Plugins.LCGImpl.os')
    def testWin(self, mock_os, mock_Popen, localSize=9001, remoteSize=9001):

        # copy a file and have it succeed
        testObject = getPluginObject(mock_os, mock_Popen, localSize, remoteSize)

        # Do transfer, match local<->remote file sizes and have it work
        newPfn = testObject.doTransfer('file:///store/NONEXISTANTSOURCE',
                              'srm://nonexistant.com/blah/?SFN=/store/NONEXISTANTTARGET',
                              True,
                              None,
                              None,
                              None,
                              None,
                              None)
        self.assertEqual(newPfn, 'srm://nonexistant.com/blah/?SFN=/store/NONEXISTANTTARGET')

if __name__ == "__main__":
    unittest.main()



#!/bin/env python



from __future__ import division
from builtins import object

import os
import os.path
import threading
import logging
import time
import base64

from subprocess import Popen, PIPE

from WMCore.WMBS.JobGroup     import JobGroup
from WMCore.WMBS.Job          import Job
from WMCore.DAOFactory        import DAOFactory

from WMCore.WMSpec.WMWorkload               import WMWorkload, WMWorkloadHelper
from WMCore.WMSpec.WMTask                   import WMTask, WMTaskHelper

#This whole thing is meant to be executed in a thread
#Thus it assumes one jobGroup per instance
#I'm wondering if it's better to multithread in a different way


class CreateScript(object):
    """
    Simple class for creating file objects

    """

    def __init__(self, name = 'default', shell = 'bash'):
        self.name = name
        self.script = """
"""
        self.shell = ''


    def setName(self, name):
        """
        Change the file name
        """
        self.name = name
        return

    def setShell(self, shell):
        """
        Set the environment shell
        """
        self.shell = shell

    def getEnv(self):
        """
        Get environment from SHELL
        """

        self.shell = os.getenv('SHELL').split('/')[-1]
        return

    def startScript(self):
        if   self.shell == 'bash':
            self.addLine('#!/bin/bash')
        elif self.shall == 'csh':
            self.addLine('#!/bin/csh')
        else:
            self.addLine('#!/bin/sh')

    def addLine(self, line = '\n'):
        """
        Add line of to the script
        """
        self.script += line + '\n'
        return

    def addEnv(self, var, value):
        """
        Add a shell-specific environment value
        This automatically appends
        """
        if os.getenv(var) == None:
            if self.shell == 'csh':
                self.addLine('export %s %s' %(var, value))
            else:
                self.addLine('export %s=%s' %(var, value))
        else:
            if self.shell == 'csh':
                self.addLine('export %s $%s:%s' %(var, var, value))
            else:
                self.addLine('export %s=$%s:%s' %(var, var, value))


    def save(self):
        """
        Now save it to the given name
        """
        if os.path.exists(self.name):
            msg = 'I will not overwrite file %s' %(self.name)
            raise Exception(msg)
        with open(self.name, 'w') as fd:
            fd.write(self.script)
        return

class CreateWorkArea(object):
    """
    Basic class for doing the JobMaker dirty work

    """



    def __init__(self, jobGroupID = None, startDir = None):


        myThread           = threading.currentThread()
        myThread.logger    = logging.getLogger()

        self.jobGroupID    = jobGroupID
        self.jobGroup      = None
        self.subscript     = None
        self.workflow      = None
        self.collectionDir = None
        self.wmWorkload    = None
        if not startDir:
            self.startDir     = os.getcwd()
        else:
            self.startDir  = startDir

        self.jobs          = {}

        self.getNewJobGroup(jobGroupID)

        self.timing = {'getNewJobGroup': 0, 'createJobGroupArea': 0, 'createWorkArea': 0, 'createDirectories': 0, 'cacheNaming': 0}

        return

    def reset(self):
        """
        Reset key variables between runs

        """

        self.workflow   = None
        self.wmWorkload = None


    def processJobs(self, jobGroup, startDir, wmWorkload = None, workflow = None):
        """
        Process the work

        This allows you to pass in two pre-loaded objects, the WMWorkloadSpace and the
        WMBS workflow, to save loading time
        """
        self.reset()
        self.wmWorkload = wmWorkload
        self.workflow   = workflow
        self.getNewJobGroup(jobGroup = jobGroup, startDir = startDir)
        self.createJobGroupArea()
        self.createWorkArea()

        self.timing = {'getNewJobGroup': 0, 'createJobGroupArea': 0, 'createWorkArea': 0, 'createDirectories': 0, 'cacheNaming': 0}

        return



    def getNewJobGroup(self, jobGroup = None, startDir = None):
        """
        This gets a job group passed to the thread
        """

        #See if we actually have a jobGroupID
        if jobGroup:
            self.jobGroupID = jobGroup.id
        else:
            #Then we have no jobGroup
            return

        if startDir:
            self.startDir = startDir

        myThread = threading.currentThread()


        #We need the workflow to get the spec
        if self.workflow == None:
            # If we have something in the workflow,
            # assume we were passed a loaded workflow
            # We need the subscription mostly to get the workflow
            self.subscript = jobGroup.subscription
            self.subscript.load()
            self.workflow  = self.subscript['workflow']
            self.workflow.load()

        if not jobGroup.exists():
            msg = 'JobMaker: Was passed a non-existant Job Group ID %i' %(self.jobGroupID)
            logging.error(msg)
            raise Exception(msg)

        self.jobGroup = jobGroup

        return




    def createJobGroupArea(self):
        """
        Creates an area for the task which is just the jobGroupUID, in which the jobs will be put

        """

        workloadDir, taskDir = self.getMasterName()

        #Create the workload directory
        if not os.path.isdir(workloadDir):
            os.makedirs(workloadDir)

        #Create the task directory
        if not os.path.isdir(taskDir):
            os.makedirs(taskDir)

        logging.info('JobMaker: Now in directory %s' %(os.getcwd()))

        return




    def createWorkArea(self):
        """
        This should handle the master tasks of creating a working area
        It should take a valid jobGroup and call the functions that create the components

        """
        myThread = threading.currentThread()

        if self.jobGroup == None:
            msg = 'JobMaker: Ordered to create for non-existant jobGroup  Failing'
            logging.error(msg)
            raise Exception(msg)

        workloadDir, taskDir       = self.getMasterName()
        jobCounter    = 0
        jobCollDir    = 0
        nameList      = []

        factory = DAOFactory("WMCore.WMBS", myThread.logger, myThread.dbi)
        setBulkCache=factory(classname = "Jobs.SetCache")
        nameDictList = []


        #Now actually start to do things
        myThread.transaction.begin()
        #jobList = self.jobGroup.listJobIDs()
        for job in self.jobGroup.jobs:
            jid = job['id']

            if jobCounter%1000 == 0:
                #Create a new jobCollection
                #Increment jobCreator if there's already something there
                jobCounter += self.createJobCollection(jobCounter, taskDir)

            jobCounter = jobCounter + 1

            name = self.getDirectoryName(jid)
            nameList.append(name)
            nameDictList.append({'jobid':jid, 'cacheDir':name})
            job['cache_dir'] = name


        setBulkCache.execute(jobDictList = nameDictList)

        self.createDirectories(nameList)

        myThread.transaction.commit()

        return




    def createJobCollection(self, jobCounter, taskDir):
        """

        Create a sub-directory to allow storage of large jobs
        """

        value = jobCounter // 1000
        jobCollDir = '%s/JobCollection_%i_%i' %(taskDir, self.jobGroup.id, value)
        #Set this to a global variable
        self.collectionDir = jobCollDir
        if not os.path.exists(jobCollDir):
            #This should be the only application
            #You return 0 because the directory you just made should be empty
            os.mkdir(jobCollDir)
            return 0
        if os.path.isdir(jobCollDir):
            #This should never happen
            return len(os.listdir(jobCollDir))
        elif os.path.isfile(jobCollDir):
            #Well, you're screwed.  Some other file is in the way: IN A DIRECTORY YOU JUST CREATED.
            #Time to freak the hell out
            raise Exception ("Could not create jobCollection %s; non-directory file in the way!" %(jobCollDir))
        else:
            #You're screwed
            raise Exception ('There was something in the way at %s, but we could not determine what it was' %(jobCollDir))



    def createDirectories(self, dirList):
        """
        Create the directory if everything is sane

        """
        myThread = threading.currentThread()


        #This is gonna be tricky
        cmdList = []
        cmdArgs = ['mkdir']

        while len(dirList) > 500:
            cmdArgs.extend(dirList[:500])
            cmdList.append(cmdArgs)
            cmdArgs = ['mkdir']
            dirList = dirList[500:]
        if len(dirList) > 0:
            cmdArgs.extend(dirList)
            cmdList.append(cmdArgs)

        logging.info('Executing makedir commands')
        for command in cmdList:
            pipe = Popen(command, stdout = PIPE, stderr = PIPE, shell = False)
            pipe.wait()


        return




    def createLocalBin(self, workdir, binName = 'localBin'):
        """
        Creates a localBin in the working directory

        """

        myThread = threading.currentThread()

        binPath = workdir + '/' + binName

        if not os.path.isdir(workdir):
            msg = 'JobMaker: Ordered to generate environment script in non-existant directory %s' %(workdir)
            logging.error(msg)
            raise Exception(msg)

        if os.path.exists(binPath):
            #This is probably not a critical error
            logging.error('Someone already created %s' %(binPath))
            return

        os.mkdir(binPath)

        return



    def generateRunScript(self, workdir, scriptName = 'run.sh'):
        """
        This generates a basic run script in the workdir directory

        """

        myThread = threading.currentThread()

        scriptPath = workdir+'/'+scriptName

        if not os.path.isdir(workdir):
            msg = 'JobMaker: Ordered to generate run script in non-existant directory %s' %(workdir)
            logging.error(msg)
            raise Exception(msg)

        RunScript = CreateScript(scriptPath)

        RunScript.getEnv()
        RunScript.startScript()
        RunScript.addLine('export PRODAGENT_JOBSPEC=$1')
        RunScript.addLine('echo \"Job Spec: $PRODAGENT_JOBSPEC\"')
        RunScript.addLine(". jobEnvironment.sh")
        RunScript.addLine("shreek --config=./ShREEKConfig.xml")

        RunScript.save()


        return





    def generateEnvironmentScript(self, workdir, scriptName = 'jobEnvironment.sh'):
        """
        This generates a basic environment sourcing script in the workdir directory

        """

        myThread = threading.currentThread()

        scriptPath = workdir+'/'+scriptName

        if not os.path.isdir(workdir):
            msg = 'JobMaker: Ordered to generate environment script in non-existant directory %s' %(workdir)
            logging.error(msg)
            raise Exception(msg)

        EnvScript = CreateScript(scriptPath)

        EnvScript.getEnv()
        EnvScript.startScript()
        EnvScript.addEnv('PYTHONPATH','`pwd`/localPython')
        EnvScript.addEnv('PATH','`pwd`/localBin')
        EnvScript.addEnv('PRODAGENT_JOB_DIR','`pwd`')
        #EnvScript.addEnv('PRODAGENT_WORKFLOW_SPEC',"`pwd`/%s" % os.path.basename(self.workflowFile))

        EnvScript.save()





    def getDirectoryName(self, jid):
        """
        Gets a universal name for the directory we're working in.

        """

        name = 'job_%i' %(jid)

        return os.path.join(self.collectionDir, name)

    def getMasterName(self):
        """
        Gets a universal name for the jobGroup directory
        Return the uid as the name if none available (THIS SHOULD NEVER HAPPEN)

        """

        if self.wmWorkload != None:
            workload = self.wmWorkload.name()
        elif not os.path.exists(self.workflow.spec):
            logging.error("Could not find Workflow spec %s; labeling jobs by job ID only!" %(self.workflow.spec))
            return os.path.join(self.startDir, self.jobGroup.uid), os.path.join(self.startDir, self.jobGroup.uid)
        else:
            wmWorkload = WMWorkloadHelper(WMWorkload("workload"))
            wmWorkload.load(self.workflow.spec)

            workload = wmWorkload.name()

        task = self.workflow.task
        if task.startswith("/" + workload + "/"):
            task = task[len(workload) + 2:]

        return os.path.join(self.startDir, workload), os.path.join(self.startDir, workload, task)



#!/usr/bin/env python
"""
_DBSBufferDataset_

Module that describes a dataset in DBS buffer

Created on May 2, 2013

@author: dballest
"""

from WMCore.WMConnectionBase import WMConnectionBase

class DBSBufferDataset(WMConnectionBase, dict):
    """
    DBSBuffer dataset object, it also contains subscription
    information
    """

    def __init__(self, path, id = -1,
                 processingVer = 0, acquisitionEra = None, validStatus = None,
                 globalTag = None, parent = None, prep_id = None):
        """
        Initialize the stored attributes and
        database connection.
        """
        WMConnectionBase.__init__(self, daoPackage = "WMComponent.DBS3Buffer")

        # Fill out the attributes
        self['path'] = path
        self['id'] = id
        self['processingVer'] = processingVer
        self['acquisitionEra'] = acquisitionEra
        self['validStatus'] = validStatus
        self['globalTag'] = globalTag
        self['parent'] = parent
        self['prep_id'] = prep_id
        self['subscriptions'] = []

    def exists(self):
        """
        _exists_

        Determine whether or not a dataset with this path exists inside the
        database.  Return the datasets's ID if it exists, False otherwise.
        """
        action = self.daofactory(classname = "ListDataset")
        result = action.execute(datasetPath = self["path"],
                                conn = self.getDBConn(),
                                transaction = self.existingTransaction())
        if not result:
            return False
        self['id'] = result[0]['id']
        return self['id']

    def create(self):
        """
        _create_

        Create a new dataset with the information stored in the object.
        Return it's ID.
        """
        if self.exists():
            self.load()
        else:
            action = self.daofactory(classname = "NewDataset")
            action.execute(datasetPath = self['path'],
                           acquisitionEra = self['acquisitionEra'],
                           processingVer = self['processingVer'],
                           validStatus = self['validStatus'],
                           globalTag = self['globalTag'],
                           parent = self['parent'],
                           prep_id = self['prep_id'],
                           conn = self.getDBConn(), transaction = self.existingTransaction())
        return self.exists()

    def load(self, subscriptions = True):
        """
        _load_

        Load a the dataset information out the database
        with subscription information included
        """
        action = self.daofactory(classname = "ListDataset")
        result = action.execute(datasetPath = self["path"],
                                    conn = self.getDBConn(),
                                    transaction = self.existingTransaction())
        if result:
            self.update(result[0])
        else:
            raise RuntimeError("No dataset present with path %s" % self['path'])
        if subscriptions:
            self.loadSubscriptions()
        return

    def updateDataset(self):
        """
        _updateDataset_

        Update the database record for this dataset with
        the information contained in this object, it must exist
        """
        if self['id'] == -1:
            raise RuntimeError("Dataset doesn't exist in the database")
        action = self.daofactory(classname = "UpdateDataset")
        action.execute(self["id"],
                       acquisitionEra = self['acquisitionEra'],
                       processingVer = self['processingVer'],
                       validStatus = self['validStatus'],
                       globalTag = self['globalTag'],
                       parent = self['parent'],
                       prep_id = self['prep_id'],
                       conn = self.getDBConn(), transaction = self.existingTransaction())
        return

    def loadSubscriptions(self):
        """
        _loadSubscriptions_

        Load subscription info out of the database
        """
        if not self.exists():
            self.load(subscriptions = False)
        action = self.daofactory(classname = "ListSubscriptions")
        result = action.execute(datasetID = self['id'],
                                conn = self.getDBConn(),
                                transaction = self.existingTransaction())
        self['subscriptions'] = result
        return

    def addSubscription(self, subscriptionInformation):
        """
        _addSubscription_

        Associate a new subscription to the dataset
        """
        if not self.exists():
            self.load(subscriptions = False)
        action = self.daofactory(classname = "NewSubscription")
        action.execute(self['id'], subscriptionInformation,
                       conn = self.getDBConn(), transaction = self.existingTransaction())

        # Refresh subscription info
        self.loadSubscriptions()
        return



#!/usr/bin/env python
"""
_Timestamps_ module designed to insert proper WM timing into WMCore FJR.
Usage: python3 Timestamps.py --reportFile=$outputFile --wmJobStart=<sec> --wmJobEnd=<sec>
where outputFile represents output FJR file, and wmJobStart/wmJobEnd represent
start and end time of WM job, respectively.

NOTE: this script should reside in Utils area of WMCore repository since it is required
at run time on condor node where there is no WMCore installation and we only provide WMCore.zip
archive, and worker node may not have unzip tool to extract this script.

"""

import getopt
import logging
import os
import pickle
import sys


# script options
options = {"reportFile=": "", "wmJobStart=": '', "wmJobEnd=": ''}


def addTimestampMetrics(config, wmJobStart, wmJobEnd):
    """
    Adjust timestamp metrics of provied FJR data.

    :param config: input FJR data
    :param wmJobStart: start time of WM job in seconds
    :param wmJobEnd: end time of WM job in seconds
    """
    wmTiming = config.section_('WMTiming')
    wmTiming.WMJobStart = wmJobStart
    wmTiming.WMJobEnd = wmJobEnd
    wmTiming.WMTotalWallClockTime = wmJobEnd - wmJobStart
    return config


def main():
    """
    Main function of the script performs business logic:
    - parse input arguments
    - read provided input FJR
    - adjust timestamp metrics of FJR data
    - write out FJR
    """
    try:
        opts, _ = getopt.getopt(sys.argv[1:], "", list(options.keys()))
    except getopt.GetoptError as ex:
        msg = "Error processing commandline args:\n"
        msg += str(ex)
        logging.error(msg)
        sys.exit(1)

    for opt, arg in opts:
        if opt == "--reportFile":
            reportFile = arg
        if opt == "--wmJobStart":
            wmJobStart = float(arg)
        if opt == "--wmJobEnd":
            wmJobEnd = float(arg)

    # read content of given FJR report file
    data = {}
    with open(reportFile, 'rb') as istream:
        data = pickle.load(istream)

    # adjust FJR data with provided metrics
    data = addTimestampMetrics(data, wmJobStart, wmJobEnd)

    # write content of FJR back to report file
    reportOutFile = reportFile + ".new"
    msg = f"Adding wmJobTime metric to {reportFile}"
    with open(reportOutFile, 'wb') as ostream:
        logging.info(msg)
        pickle.dump(data, ostream)

    # if we successfully wrote reportOutFile we can swap it with input one
    sizeStatus = os.path.getsize(reportOutFile) >= os.path.getsize(reportFile)
    if os.path.isfile(reportOutFile) and sizeStatus:
        os.rename(reportOutFile, reportFile)


if __name__ == '__main__':
    main()



#!/usr/bin/env python


"""
StatusPoller

Possible non-production poller prototype for
JobStatusAir
"""

from future.utils import listvalues

import time
import logging
import threading
from collections import defaultdict

from Utils.IteratorTools import flattenList
from Utils.Timers import timeFunction
from WMCore.WMException                    import WMException
from WMCore.WMExceptions                   import WM_JOB_ERROR_CODES
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.BossAir.BossAirAPI             import BossAirAPI

class StatusPollerException(WMException):
    """
    _StatusPollerException_

    Exception class for StatusPoller errors.
    """

class StatusPoller(BaseWorkerThread):
    """
    _StatusPoller_

    Prototype for polling for
    JobStatusAir
    """

    def __init__(self, config):
        """
        __init__

        Set up the caching and other objects
        """
        self.config = config
        BaseWorkerThread.__init__(self)

        self.cachedJobs = []

        self.bossAir = BossAirAPI(config=config)

        # With no timeouts, nothing ever happens
        # Otherwise we expect a dictionary with the keys representing
        # the states and the values the timeouts.
        self.timeouts = getattr(config.JobStatusLite, 'stateTimeouts')

        return

    @timeFunction
    def algorithm(self, parameters=None):
        """
        _algorithm_

        Handle any exceptions with the actual code
        """
        myThread = threading.currentThread()
        try:
            logging.info("Running job status poller algorithm...")
            self.checkStatus()
        except WMException as ex:
            if getattr(myThread, 'transaction', None):
                myThread.transaction.rollbackForError()
            raise
        except Exception as ex:
            msg = "Unhandled error in statusPoller"
            msg += str(ex)
            logging.exception(msg)
            if getattr(myThread, 'transaction', None):
                myThread.transaction.rollbackForError()
            raise StatusPollerException(msg)

        return

    def checkStatus(self):
        """
        _checkStatus_

        Run the BossAir track() function (self-contained)
        and then check for jobs that have timed out.
        """


        runningJobs = self.bossAir.track()

        if len(runningJobs) < 1:
            # Then we have no jobs
            return

        if not self.timeouts:
            # Then we've set ourselves to have no timeouts
            # Get out and stay out
            return

        # Look for jobs that need to be killed
        jobsToKill = defaultdict(list)

        # Now check for timeouts
        for job in runningJobs:
            globalState = job.get('globalState', 'Error')
            statusTime = job.get('status_time', None)
            timeout = self.timeouts.get(globalState, None)
            if statusTime == 0:
                logging.error("Not killing job %i, the status time was zero", job['id'])
                continue
            if timeout and statusTime:
                if time.time() - float(statusTime) > float(timeout):
                    # Timeout status is used by JobTracker to fail jobs in WMBS database
                    logging.info("Killing job %i because it has exceeded timeout for status '%s'", job['id'], globalState)
                    job['status'] = 'Timeout'
                    jobsToKill[globalState].append(job)

        timeOutCodeMap = {"Running": 71304, "Pending": 71305, "Error": 71306}
        # We need to show that the jobs are in state timeout
        # and then kill them.
        jobsToKillList = flattenList(listvalues(jobsToKill))
        myThread = threading.currentThread()
        myThread.transaction.begin()
        self.bossAir.update(jobs=jobsToKillList)
        for preJobStatus in jobsToKill:
            eCode = timeOutCodeMap.get(preJobStatus, 71307) # it shouldn't have 71307 (states should be among Running, Pending, Error)
            self.bossAir.kill(jobs=jobsToKill[preJobStatus], killMsg=WM_JOB_ERROR_CODES[eCode], errorCode=eCode)
        myThread.transaction.commit()

        return

    def terminate(self, params):
        """
        _terminate_

        Kill the code after one final pass when called by the master thread.
        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(params)



#!/usr/bin/env python
"""
_Transitions_

Controls what state transitions are allowed.
"""

from future.utils import viewvalues


class Transitions(dict):
    """
    All allowed state transitions in the JSM.
    """
    def __init__(self):
        self.setdefault('none', ['new'])
        self.setdefault('new', ['created', 'createfailed', 'killed'])
        self.setdefault('created', ['executing', 'submitfailed', 'createfailed', 'killed'])
        self.setdefault('executing', ['complete', 'jobfailed', 'killed'])
        self.setdefault('complete', ['jobfailed', 'success'])
        self.setdefault('createfailed', ['createcooloff', 'retrydone', 'killed'])
        self.setdefault('submitfailed', ['submitcooloff', 'retrydone', 'killed'])
        self.setdefault('jobfailed', ['jobcooloff', 'retrydone', 'killed'])
        self.setdefault('createcooloff', ['created', 'createpaused', 'retrydone', 'killed'])
        self.setdefault('submitcooloff', ['created', 'submitpaused', 'retrydone', 'killed'])
        self.setdefault('jobcooloff', ['created', 'jobpaused', 'retrydone', 'killed'])
        self.setdefault('success', ['cleanout'])
        self.setdefault('retrydone', ['exhausted', 'killed'])
        self.setdefault('exhausted', ['cleanout'])
        self.setdefault('killed', ['cleanout', 'killed'])
        self.setdefault('jobpaused', ['created', 'retrydone', 'killed'])
        self.setdefault('createpaused', ['created', 'retrydone', 'killed'])
        self.setdefault('submitpaused', ['created', 'retrydone', 'killed'])

    def states(self):
        """
        _states_

        Return a list of all known states, derive it in case we add new final
        states other than cleanout.
        """
        knownstates = set(self.keys())
        for possiblestates in viewvalues(self):
            for i in possiblestates:
                knownstates.add(i)
        return list(knownstates)



#!/usr/bin/env python

"""
This code should load the necessary information regarding
dataset-algo combinations from the DBSBuffer.

"""




from WMCore.Database.DBFormatter import DBFormatter


class FindDASToUpload(DBFormatter):
    """
    Find Uploadable DAS

    """
    sql = """SELECT dbsbuffer_dataset.path AS datasetpath,
                    dbsbuffer_dataset.acquisition_era AS acquera,
                    dbsbuffer_dataset.processing_ver AS procver
             FROM dbsbuffer_file
             INNER JOIN dbsbuffer_algo_dataset_assoc ON
               dbsbuffer_algo_dataset_assoc.id = dbsbuffer_file.dataset_algo
             INNER JOIN dbsbuffer_dataset ON
               dbsbuffer_dataset.id = dbsbuffer_algo_dataset_assoc.dataset_id
             LEFT OUTER JOIN dbsbuffer_file_parent ON
               dbsbuffer_file_parent.child = dbsbuffer_file.id
             LEFT OUTER JOIN dbsbuffer_file parent_file ON
               parent_file.id = dbsbuffer_file_parent.parent AND
               parent_file.status = 'NOTUPLOADED'
             WHERE dbsbuffer_file.status = 'NOTUPLOADED'
             GROUP BY dbsbuffer_dataset.path,
                      dbsbuffer_dataset.acquisition_era,
                      dbsbuffer_dataset.processing_ver
             HAVING COUNT(parent_file.id) = 0
             """

    def makeDAS(self, results):
        """
        build the data structure we want to return

        """
        datasetalgos = []
        for result in results:
            datasetalgos.append( { 'DatasetPath' : result['datasetpath'],
                                   'AcquisitionEra' : result['acquera'],
                                   'ProcessingVer' : result['procver'] } )

        return datasetalgos

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn, transaction = transaction)
        return self.makeDAS(self.formatDict(result))



#!/usr/bin/env python
"""
_RunBased_t_

RunBased splitting test.
"""
from __future__ import division

from builtins import range

import unittest

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.Job import Job
from WMCore.DataStructs.Subscription import Subscription
from WMCore.DataStructs.Workflow import Workflow
from WMCore.DataStructs.Run import Run

from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID

class RunBasedTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """
        self.multipleFileFileset = Fileset(name = "TestFileset1")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100, locations = set(["somese.cern.ch"]))
            newFile.addRun(Run(i, *[45+i]))
            self.multipleFileFileset.addFile(newFile)

        self.singleFileFileset = Fileset(name = "TestFileset2")
        newFile = File("/some/file/name", size = 1000, events = 100, locations = set(["somese.cern.ch"]))
        newFile.addRun(Run(1, *[45]))
        self.singleFileFileset.addFile(newFile)


        self.multipleFileRunset = Fileset(name = "TestFileset3")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100, locations = set(["somese.cern.ch"]))
            newFile.addRun(Run(i//3, *[45]))
            self.multipleFileRunset.addFile(newFile)

        self.singleRunFileset = Fileset(name = "TestFileset4")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100, locations = set(["somese.cern.ch"]))
            newFile.addRun(Run(1, *[45]))
            self.singleRunFileset.addFile(newFile)

        testWorkflow = Workflow()
        self.multipleFileSubscription = Subscription(fileset = self.multipleFileFileset,
                                                     workflow = testWorkflow,
                                                     split_algo = "RunBased",
                                                     type = "Processing")
        self.singleFileSubscription   = Subscription(fileset = self.singleFileFileset,
                                                     workflow = testWorkflow,
                                                     split_algo = "RunBased",
                                                     type = "Processing")
        self.multipleRunSubscription  = Subscription(fileset = self.multipleFileRunset,
                                                     workflow = testWorkflow,
                                                     split_algo = "RunBased",
                                                     type = "Processing")
        self.singleRunSubscription    = Subscription(fileset = self.singleRunFileset,
                                                     workflow = testWorkflow,
                                                     split_algo = "RunBased",
                                                     type = "Processing")


        return

    def tearDown(self):
        pass

    def testExactRuns(self):
        """
        _testExactRuns_

        Test run based job splitting when the number of events per job is
        exactly the same as the number of events in the input file.
        """

        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        jobGroups = jobFactory(files_per_job = 1)

        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't create a single job."

        job = jobGroups[0].jobs.pop()

        assert job.getFiles(type = "lfn") == ["/some/file/name"], \
               "ERROR: Job contains unknown files."

        return


    def testMoreRuns(self):
        """
        _testMoreEvents_

        Test run based job splitting when the number of runs per job is
        greater than the number of runs in the input file.
        """

        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        jobGroups = jobFactory(files_per_job = 2)

        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't create a single job."

        job = jobGroups[0].jobs.pop()

        assert job.getFiles(type = "lfn") == ["/some/file/name"], \
               "ERROR: Job contains unknown files."

        return

    def testMultipleRuns(self):
        """
        _testMultipleRuns_

        Test run based job splitting when the number of runs is
        equal to the number in each input file, with multiple files

        """

        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        jobGroups = jobFactory(files_per_job = 1)

        assert len(jobGroups) == 10, \
               "ERROR: JobFactory didn't return one JobGroup per run."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't put each run in a file."

        self.assertEqual(len(jobGroups[0].jobs.pop().getFiles(type = "lfn")), 1)


        return

    def testMultipleRunsCombine(self):
        """
        _testMultipleRunsCombine_

        Test run based job splitting when the number of jobs is
        less then the number of files, with multiple files

        """

        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleRunSubscription)

        jobGroups = jobFactory(files_per_job = 2)



        assert len(jobGroups) == 4, \
               "ERROR: JobFactory didn't return one JobGroup per run."

        assert len(jobGroups[1].jobs) == 2, \
               "ERROR: JobFactory didn't put only one job in the first job"

        #Last one in the queue should have one job, previous two (three files per run)
        self.assertEqual(len(jobGroups[1].jobs.pop().getFiles(type = "lfn")), 1)
        self.assertEqual(len(jobGroups[1].jobs.pop().getFiles(type = "lfn")), 2)


        return

    def testSingleRunsCombineUneven(self):
        """
        _testSingleRunsCombineUneven_

        Test run based job splitting when the number of jobs is
        less then and indivisible by the number of files, with multiple files.

        """

        #This should return two jobs, one with 8 and one with 2 files

        splitter = SplitterFactory()
        jobFactory = splitter(self.singleRunSubscription)

        jobGroups = jobFactory(files_per_job = 8)

        self.assertEqual(len(jobGroups),         1)
        self.assertEqual(len(jobGroups[0].jobs), 2)
        self.assertEqual(len(jobGroups[0].jobs.pop().getFiles(type = "lfn")), 2)
        self.assertEqual(len(jobGroups[0].jobs.pop().getFiles(type = "lfn")), 8)


        return




if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_FileCountBySubscription_

Monitoring DAO classes for Jobs in WMBS
TODO: this is not working version
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.DefaultFormatter import DefaultFormatter

class FileCountBySubscriptionAndRun(DefaultFormatter):
    """
    _FileCountBySubscription_

    count the number of files for each state for given subscription and run
    """

    sql = """SELECT COUNT(DISTINCT wmbs_sub_files_acquired.fileid) AS acquired
               ,COUNT(DISTINCT wmbs_sub_files_complete.fileid) AS complete
                ,COUNT(DISTINCT wmbs_sub_files_failed.fileid) AS failed
              FROM wmbs_file_details
              INNER JOIN wmbs_file_dataset_path_assoc ON wmbs_file_dataset_path_assoc.file_id = wmbs_file_details.id
              INNER JOIN dataset_path ON wmbs_file_dataset_path_assoc.dataset_path_id = dataset_path.id
              INNER JOIN data_tier ON dataset_path.data_tier = data_tier.id
              INNER JOIN wmbs_file_runlumi_map ON wmbs_file_details.id = wmbs_file_runlumi_map.fileid
              LEFT OUTER JOIN wmbs_sub_files_acquired ON wmbs_file_details.id = wmbs_sub_files_acquired.fileid
              LEFT OUTER JOIN wmbs_sub_files_complete ON wmbs_file_details.id = wmbs_sub_files_complete.fileid
              LEFT OUTER JOIN wmbs_sub_files_failed ON wmbs_file_details.id = wmbs_sub_files_failed.fileid
              WHERE run = :run
              AND (wmbs_sub_files_acquired.subscription = (SELECT id FROM wmbs_subscription WHERE workflow =
              (SELECT id FROM wmbs_workflow WHERE spec = 'FileMerge'))
              OR wmbs_sub_files_complete.subscription =
              (SELECT id FROM wmbs_subscription WHERE workflow =
              (SELECT id FROM wmbs_workflow WHERE spec = 'FileMerge'))
              OR wmbs_sub_files_failed.subscription =
              (SELECT id FROM wmbs_subscription WHERE workflow =
              (SELECT id FROM wmbs_workflow WHERE spec = 'FileMerge')))
               AND data_tier.name = 'RECO'
            """

    def execute(self, fileset_name, workflow_name, run, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        bindVars = {"fileset_name": fileset_name, "workflow_name": workflow_name, "run": run}

        result = self.dbi.processData(self.sql, bindVars, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



# system modules
import cherrypy
from cheroot.test import webtest
from cherrypy import expose
from multiprocessing import Process

# WMCore modules
from WMCore.REST.Auth import user_info_from_headers
from WMCore.REST.Test import setup_dummy_server, fake_authz_headers
from WMCore.REST.Test import fake_authz_key_file
from WMCore.REST.Tools import tools

FAKE_FILE = fake_authz_key_file()
PORT = 8888

class Root:
    def __init__(self, *args):
        pass

    @expose
    def default(self):
        return "foo"

    @expose
    @tools.cms_auth(role = "Global Admin", group = "global")
    def global_admin(self):
        return "ok"

class SimpleTest(webtest.WebCase):

    def setUp(self):
        self.h = fake_authz_headers(FAKE_FILE.data)
        self.hglobal = fake_authz_headers(FAKE_FILE.data, roles = {"Global Admin": {'group': ['global']}})
        webtest.WebCase.PORT = PORT
        self.engine = cherrypy.engine
        self.proc = load_server(self.engine)

    def tearDown(self):
        stop_server(self.proc, self.engine)

    def test_basic_fail(self):
        self.getPage("/test")
        # we changed behaviour of REST server to bypass authz checks when RemoteAddr is localhost
        # see src/python/WMCore/REST/Auth.py, therefore this test should be ok
        # self.assertStatus("403 Forbidden")
        self.assertStatus("200 OK")

    def test_basic_success(self):
        self.getPage("/test", headers = self.h)
        self.assertStatus("200 OK")
        self.assertBody("foo")

    def test_auth_fail(self):
        self.getPage("/test/global_admin", headers = self.h)
        self.assertStatus("403 Forbidden")

    def test_auth_success(self):
        self.getPage("/test/global_admin", headers = self.hglobal)
        self.assertStatus("200 OK")
        self.assertBody("ok")

class AuthTest(webtest.WebCase, cherrypy.Tool):

    def setUp(self):
        cherrypy.Tool.__init__(self, 'before_request_body', user_info_from_headers, priority=60)
        webtest.WebCase.PORT = PORT
        self.engine = cherrypy.engine
        self.proc = load_server(self.engine)

    def tearDown(self):
        print("teardown")
        stop_server(self.proc, self.engine)

    def testAuth(self):
        myHeaders = [('cms-authn-name', 'Blah'), ('cms-auth-status', 'OK'),
                     ('cms-authn-login', 'blah'), ('cms-authn-hmac', '1234')]
        self.getPage("/test", headers=myHeaders)
        self.assertTrue(True)  # Do not remove this line! otherwise the test hangs


def setup_server():
    srcfile = __file__.split("/")[-1].split(".py")[0]
    setup_dummy_server(srcfile, "Root", authz_key_file=FAKE_FILE, port=PORT)

def load_server(engine):
    setup_server()
    proc = Process(target=start_server, name="cherrypy_Api_t", args=(engine,))
    proc.start()
    proc.join(timeout=1)
    return proc

def start_server(engine):
    webtest.WebCase.PORT = PORT
    cherrypy.log.screen = True
    engine.start()
    engine.block()

def stop_server(proc, engine):
    cherrypy.log.screen = True
    engine.stop()
    proc.terminate()

if __name__ == '__main__':
    webtest.main()



#!/usr/bin/env python
"""
_GetCompletedFiles_

Oracle implementation of Subscription.GetCompletedFiles
"""

from WMCore.WMBS.MySQL.Subscriptions.GetCompletedFiles import \
     GetCompletedFiles as GetCompletedFilesMySQL

class GetCompletedFiles(GetCompletedFilesMySQL):
    pass



#!/usr/bin/env python
"""
_SetBlock_

Oracle implementation of DBSBufferFiles.SetBlock
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.SetBlock import SetBlock as MySQLSetBlock




class SetBlock(MySQLSetBlock):
    pass
